#! /usr/bin/env python
from math import pow

from solid import union, circle, cylinder, polygon, color, OpenSCADObject, translate, linear_extrude, polyhedron
from solid.utils import bounding_box, right, Red, Tuple3, euclidify
from euclid3 import Vector2, Vector3, Point2, Point3

from typing import Sequence, Tuple, Union, List, cast

Point23 = Union[Point2, Point3]
# These *Input types accept either euclid3.Point* objects, or bare n-tuples
Point2Input = Union[Point2, Tuple[float, float]]
Point3Input = Union[Point3, Tuple[float, float, float]]
Point23Input = Union[Point2Input, Point3Input]

PointInputs = Sequence[Point23Input]

FaceTrio = Tuple[int, int, int]
CMPatchPoints = Tuple[Sequence[Point3Input], Sequence[Point3Input]]

Vec23 = Union[Vector2, Vector3]
FourPoints = Tuple[Point23Input, Point23Input, Point23Input, Point23Input]
SEGMENTS = 48

DEFAULT_SUBDIVISIONS = 10
DEFAULT_EXTRUDE_HEIGHT = 1

# =======================
# = CATMULL-ROM SPLINES =
# =======================
def catmull_rom_polygon(points: Sequence[Point23Input], 
                        subdivisions: int = DEFAULT_SUBDIVISIONS, 
                        extrude_height: float = DEFAULT_EXTRUDE_HEIGHT, 
                        show_controls: bool =False,
                        center: bool=True) -> OpenSCADObject:
    """
    Return a closed OpenSCAD polygon object through all of `points`, 
    extruded to `extrude_height`. If `show_controls` is True, return red 
    cylinders at each of the specified control points; this makes it easier to
    move determine which points should move to get a desired shape.

    NOTE: if `extrude_height` is 0, this function returns a 2D `polygon()`
    object, which OpenSCAD can only combine with other 2D objects 
    (e.g. `square`, `circle`, but not `cube` or `cylinder`). If `extrude_height`
    is nonzero, the object returned will be 3D and only combine with 3D objects.
    """
    catmull_points = catmull_rom_points(points, subdivisions, close_loop=True)
    shape = polygon(catmull_points)
    if extrude_height > 0:
        shape = linear_extrude(height=extrude_height, center=center)(shape)

    if show_controls:
        shape += control_points(points, extrude_height, center)
    return shape

def catmull_rom_points( points: Sequence[Point23Input], 
                        subdivisions:int = DEFAULT_SUBDIVISIONS, 
                        close_loop: bool=False,
                        start_tangent: Vec23 = None,
                        end_tangent: Vec23 = None) -> List[Point3]:
    """
    Return a smooth set of points through `points`, with `subdivisions` points 
    between each pair of control points. 
    
    If `close_loop` is False, `start_tangent` and `end_tangent` can specify 
    tangents at the open ends of the returned curve. If not supplied, tangents 
    will be colinear with first and last supplied segments

    Credit due: Largely taken from C# code at: 
    https://www.habrador.com/tutorials/interpolation/1-catmull-rom-splines/
    retrieved 20190712
    """
    catmull_points: List[Point3] = []
    cat_points: List[Point3] = []
    # points_list = cast(List[Point23], points)

    points_list = list([euclidify(p, Point3) for p in points])

    if close_loop:
        cat_points = euclidify([points_list[-1]] + points_list + points_list[0:2], Point3)
    else:
        # Use supplied tangents or just continue the ends of the supplied points
        start_tangent = start_tangent or (points_list[1] - points_list[0])
        start_tangent = euclidify(start_tangent, Vector3)
        end_tangent = end_tangent or (points_list[-2] - points_list[-1])
        end_tangent = euclidify(end_tangent, Vector3)
        cat_points = [points_list[0]+ start_tangent] + points_list + [points_list[-1] + end_tangent]

    last_point_range = len(cat_points) - 3 if close_loop else len(cat_points) - 3

    for i in range(0, last_point_range):
        include_last = True if i == last_point_range - 1 else False
        controls = cat_points[i:i+4]
        # If we're closing a loop, controls needs to wrap around the end of the array
        points_needed = 4 - len(controls)
        if points_needed > 0:
            controls += cat_points[0:points_needed]
        controls_tuple = cast(FourPoints, controls)
        catmull_points += _catmull_rom_segment(controls_tuple, subdivisions, include_last)

    return catmull_points

def _catmull_rom_segment(controls: FourPoints, 
                         subdivisions: int, 
                         include_last=False) -> List[Point3]: 
    """
    Returns `subdivisions` Points between the 2nd & 3rd elements of `controls`,
    on a quadratic curve that passes through all 4 control points.
    If `include_last` is True, return `subdivisions` + 1 points, the last being
    controls[2]. 

    No reason to call this unless you're trying to do something very specific
    """
    pos: Point23 = None
    positions: List[Point23] = []

    num_points = subdivisions
    if include_last:
        num_points += 1

    p0, p1, p2, p3 = [euclidify(p, Point3) for p in controls]
    a = 2 * p1
    b = p2 - p0
    c = 2* p0 - 5*p1 + 4*p2 - p3
    d = -p0 + 3*p1 - 3*p2 + p3

    for i in range(num_points):
        t = i/subdivisions
        pos = 0.5 * (a + (b * t) + (c * t * t) + (d * t * t * t))
        positions.append(Point3(*pos))
    return positions

def catmull_rom_patch_points(patch:Tuple[PointInputs, PointInputs], 
                             subdivisions:int = DEFAULT_SUBDIVISIONS,
                             index_start:int = 0) -> Tuple[List[Point3], List[FaceTrio]]:
    verts: List[Point3] = []
    faces: List[FaceTrio] = []

    cm_points_a = catmull_rom_points(patch[0], subdivisions=subdivisions)
    cm_points_b = catmull_rom_points(patch[1], subdivisions=subdivisions)

    strip_length = len(cm_points_a)

    for i in range(subdivisions + 1):
        frac = i/subdivisions
        verts += list([affine_combination(a,b, frac) for a,b in zip(cm_points_a, cm_points_b)])
        a_start = i*strip_length + index_start
        b_start = a_start + strip_length
        # This connects the verts we just created to the verts we'll make on the
        # next loop. So don't calculate for the last loop
        if i < subdivisions:
            faces += face_strip_list(a_start, b_start, strip_length)

    return verts, faces

def catmull_rom_patch(patch:Tuple[PointInputs, PointInputs], subdivisions:int = DEFAULT_SUBDIVISIONS) -> OpenSCADObject:

    faces, vertices = catmull_rom_patch_points(patch, subdivisions)
    return polyhedron(faces, vertices)

def catmull_rom_prism(  control_curves:Sequence[PointInputs], 
                        subdivisions:int = DEFAULT_SUBDIVISIONS,
                        closed_ring:bool = True,
                        add_caps:bool = True,
                        smooth_edges: bool = False ) -> polyhedron:
    if smooth_edges:
        return catmull_rom_prism_smooth_edges(control_curves, subdivisions, closed_ring, add_caps)

    verts: List[Point3] = []
    faces: List[FaceTrio] = []

    curves = list([euclidify(c) for c in control_curves])
    if closed_ring:
        curves.append(curves[0])
    
    curve_length = (len(curves[0]) -1) * subdivisions + 1
    for i, (a, b) in enumerate(zip(curves[:-1], curves[1:])):
        index_start = len(verts) - curve_length
        first_new_vert = curve_length
        if i == 0:
            index_start = 0
            first_new_vert = 0

        new_verts, new_faces = catmull_rom_patch_points((a,b), subdivisions=subdivisions, index_start=index_start)

        # new_faces describes all the triangles in the patch we just computed,
        # but new_verts shares its first curve_length vertices with the last
        # curve_length vertices; Add on only the new points
        verts += new_verts[first_new_vert:]
        faces += new_faces

    if closed_ring and add_caps:
        bot_indices = range(0, len(verts), curve_length)
        top_indices = range(curve_length-1, len(verts), curve_length)

        bot_centroid, bot_faces = centroid_endcap(verts, bot_indices)
        verts.append(bot_centroid)
        faces += bot_faces
        # Note that bot_centroid must be added to verts before creating the
        # top endcap; otherwise both endcaps would point to the same centroid point
        top_centroid, top_faces = centroid_endcap(verts, top_indices, invert=True)
        verts.append(top_centroid)
        faces += top_faces
    
    p = polyhedron(faces=faces, points=verts, convexity=3)
    return p

def catmull_rom_prism_smooth_edges( control_curves:Sequence[PointInputs], 
                                    subdivisions:int = DEFAULT_SUBDIVISIONS,
                                    closed_ring:bool = True,
                                    add_caps:bool = True ) -> polyhedron:

    verts: List[Point3] = []
    faces: List[FaceTrio] = []

    # TODO: verify that each control_curve has the same length

    curves = list([euclidify(c) for c in control_curves])

    expanded_curves = [catmull_rom_points(c, subdivisions, close_loop=False) for c in curves]
    expanded_length = len(expanded_curves[0])
    for i in range(expanded_length):
        contour_controls = [c[i] for c in expanded_curves]
        contour = catmull_rom_points(contour_controls, subdivisions, close_loop=closed_ring)
        verts += contour

        contour_length = len(contour)
        # generate the face triangles between the last two rows of vertices
        if i > 0:
            a_start = len(verts) - 2 * contour_length
            b_start = len(verts) - contour_length
            # Note the b_start, a_start order here. This makes sure our faces
            # are pointed outwards for the test cases I ran. I think if control
            # curves were specified clockwise rather than counter-clockwise, all
            # of the faces would be pointed inwards
            new_faces = face_strip_list(b_start,  a_start, length=contour_length, close_loop=closed_ring)
            faces += new_faces
    
    if closed_ring and add_caps:
        bot_indices = range(0, contour_length)
        top_indices = range(len(verts) - contour_length, len(verts))

        bot_centroid, bot_faces = centroid_endcap(verts, bot_indices)
        verts.append(bot_centroid)
        faces += bot_faces
        # Note that bot_centroid must be added to verts before creating the
        # top endcap; otherwise both endcaps would point to the same centroid point
        top_centroid, top_faces = centroid_endcap(verts, top_indices, invert=True)
        verts.append(top_centroid)
        faces += top_faces 

    p = polyhedron(faces=faces, points=verts, convexity=3)
    return p

# ==================
# = BEZIER SPLINES =
# ==================
# Ported from William A. Adams' Bezier OpenSCAD code at: 
# https://www.thingiverse.com/thing:8443

def bezier_polygon( controls: FourPoints, 
                    subdivisions:int = DEFAULT_SUBDIVISIONS, 
                    extrude_height:float = DEFAULT_EXTRUDE_HEIGHT,
                    show_controls: bool = False,
                    center: bool = True) -> OpenSCADObject:
    '''
    Return an OpenSCAD object representing a closed quadratic Bezier curve.
    If extrude_height == 0, return a 2D `polygon()` object. 
    If extrude_height > 0, return a 3D extrusion of specified height. 
    Note that OpenSCAD won't render 2D & 3D objects together correctly, so pick
    one and use that.
    '''                
    points = bezier_points(controls, subdivisions)
    # OpenSCAD can'ts handle Point3s in creating a polygon. Convert them to Point2s
    # Note that this prevents us from making polygons outside of the XY plane, 
    # even though a polygon could reasonably be in some other plane while remaining 2D
    points = list((Point2(p.x, p.y) for p in points))
    shape: OpenSCADObject = polygon(points)
    if extrude_height != 0:
        shape = linear_extrude(extrude_height, center=center)(shape)

    if show_controls:
        control_objs = control_points(controls, extrude_height=extrude_height, center=center)
        shape += control_objs
    
    return shape

def bezier_points(controls: FourPoints, 
                  subdivisions: int = DEFAULT_SUBDIVISIONS,
                  include_last: bool = True) -> List[Point3]:
    """
    Returns a list of `subdivisions` (+ 1, if `include_last` is True) points
    on the cubic bezier curve defined by `controls`. The curve passes through 
    controls[0] and controls[3]

    If `include_last` is True, the last point returned will be controls[3]; if
    False, (useful for linking several curves together), controls[3] won't be included

    Ported from William A. Adams' Bezier OpenSCAD code at: 
    https://www.thingiverse.com/thing:8443
    """
    # TODO: enable a smooth curve through arbitrarily many points, as described at:
    # https://www.algosome.com/articles/continuous-bezier-curve-line.html

    points: List[Point3] = []
    last_elt = 1 if include_last else 0
    for i in range(subdivisions + last_elt):
        u = i/subdivisions
        points.append(_point_along_bez4(*controls, u))
    return points

def _point_along_bez4(p0: Point23Input, p1: Point23Input, p2: Point23Input, p3: Point23Input, u:float) -> Point3:
    p0 = euclidify(p0)
    p1 = euclidify(p1)
    p2 = euclidify(p2)
    p3 = euclidify(p3)

    x = _bez03(u)*p0.x + _bez13(u)*p1.x + _bez23(u)*p2.x + _bez33(u)*p3.x
    y = _bez03(u)*p0.y + _bez13(u)*p1.y + _bez23(u)*p2.y + _bez33(u)*p3.y
    z = _bez03(u)*p0.z + _bez13(u)*p1.z + _bez23(u)*p2.z + _bez33(u)*p3.z
    return Point3(x, y, z)

def _bez03(u:float) -> float:
    return pow((1-u), 3)

def _bez13(u:float) -> float:
    return 3*u*(pow((1-u),2))

def _bez23(u:float) -> float:
    return 3*(pow(u,2))*(1-u)

def _bez33(u:float) -> float:
    return pow(u,3)

# ================
# = HOBBY CURVES =
# ================

# ===========
# = HELPERS =
# ===========
def control_points(points: Sequence[Point23], extrude_height:float=0, center:bool=True, points_color:Tuple3=Red) -> OpenSCADObject:
    """
    Return a list of red cylinders/circles (depending on `extrude_height`) at
    a supplied set of 2D points. Useful for visualizing and tweaking a curve's 
    control points
    """
    # Figure out how big the circles/cylinders should be based on the spread of points
    min_bb, max_bb = bounding_box(points)
    outline_w = max_bb[0] - min_bb[0]
    outline_h = max_bb[1] - min_bb[1]
    r = min(outline_w, outline_h) / 20 # 
    if extrude_height == 0:
        c = circle(r=r)
    else:
        h = extrude_height * 1.1
        c = cylinder(r=r, h=h, center=center)
    controls = color(points_color)([translate((p.x, p.y, 0))(c) for p in points])
    return controls

def face_strip_list(a_start:int,  b_start:int, length:int, close_loop:bool=False) -> List[FaceTrio]:
    # If a_start is the index of the vertex at one end of a row of points in a surface,
    # and b_start is the index of the vertex at the same end of the next row of points,
    # return a list of lists of indices describing faces for the whole row:
    # face_strip_list(a_start = 0, b_start = 3, length=3) => [[0,4,3], [0,1,4], [1,5,4], [1,2,5]]
    #   3-4-5
    #   |/|/|
    #   0-1-2  =>  [[0,4,3], [0,1,4], [1,5,4], [1,2,5]]
    #
    # If close_loop is true, add one more pair of faces connecting the far
    # edge of the strip to the near edge, in this case [[2,3,5], [2,0,3]]
    faces: List[FaceTrio] = []
    loop = length - 1

    for a, b in zip(range(a_start, a_start + loop), range(b_start, b_start + loop)):
        faces.append((a, b+1, b))
        faces.append((a, a+1, b+1))
    if close_loop:
        faces.append((a+loop, b, b+loop))
        faces.append((a+loop, a, b))
    return faces

def fan_endcap_list(cap_points:int=3, index_start:int=0) -> List[FaceTrio]:
    '''
    Return a face-triangles list for the endpoint of a tube with cap_points points
    We construct a fan of triangles all starting at point index_start and going
    to each point in turn. 
    
    NOTE that this would not work for non-convex rings. 
    In that case, it would probably be better to create a new centroid point and have
    all triangle reach out from it. That wouldn't handle all polygons, but would
    work with mildly concave ones like a star, for example.

    So fan_endcap_list(cap_points=6, index_start=0), like so:
           0   
         /   \
       5      1
       |      | 
       4      2
         \   /
           3      
    
           returns:  [(0,1,2), (0,2,3), (0,3,4), (0,4,5)]      
    '''
    faces: List[FaceTrio] = []
    for i in range(index_start + 1, index_start + cap_points - 1):
        faces.append((index_start, i, i+1))
    return faces

def centroid_endcap(tube_points:Sequence[Point3], indices:Sequence[int], invert:bool = False) -> Tuple[Point3, List[FaceTrio]]:
    # tube_points: all points in a polyhedron tube
    # indices: the indexes of the points at the desired end of the tube
    # invert: if True, invert the order of the generated faces. One endcap in 
    #   each pair should be inverted
    #
    # Return all the triangle information needed to make an endcap polyhedron
    #
    # This is sufficient for some moderately concave polygonal endcaps, 
    # (a star shape, say), but wouldn't be enough for more irregularly convex
    # polygons (anyplace where a segment from the centroid to a point on the 
    # polygon crosses an edge of the polygon)
    faces: List[FaceTrio] = []
    center = centroid([tube_points[i] for i in indices])
    centroid_index = len(tube_points)
    
    for a,b in zip(indices[:-1], indices[1:]):
        faces.append((centroid_index, a, b))
    faces.append((centroid_index, indices[-1], indices[0]))

    if invert:
        faces = list((reversed(f) for f in faces)) # type: ignore

    return (center, faces)

def centroid(points:Sequence[Point23]) -> Point23:
    total = Point3(0,0,0)
    for p in points:
        total += p
    total /= len(points)
    return total

def affine_combination(a:Point23, b:Point23, fraction:float) -> Point23:
    # Return a Point[23] between a & b, where fraction==0 => a, fraction==1 => b
    return (1-fraction) * a + fraction*b




#! /usr/bin/env python
import os
import sys
from solid import *
from solid.utils import Red, right, forward, back

from solid.splines import catmull_rom_points, catmull_rom_polygon, control_points
from solid.splines import bezier_polygon, bezier_points
from euclid3 import Vector2, Vector3, Point2, Point3

def assembly():
    # Catmull-Rom Splines
    a = basic_catmull_rom()                         # Top row in OpenSCAD output
    a += back(4)(catmull_rom_spline_variants())     # Row 2
    a += back(12)(bottle_shape(width=2, height=6))  # Row 3, the bottle shape

    # # TODO: include examples for 3D surfaces:
    # a += back(16)(catmull_rom_patches())
    # a += back(20)(catmull_rom_prism())
    # a += back(24)(catmull_rom_prism_smooth())

    # Bezier Splines
    a += back(16)(basic_bezier())                   # Row 4
    a += back(20)(bezier_points_variants())         # Row 5
    return a

def basic_catmull_rom():
    points = [
        Point2(0,0),
        Point2(1,1),
        Point2(2,1),
        Point2(2,-1),
    ]
    # In its simplest form, catmull_rom_polygon() will just make a C1-continuous
    # closed shape. Easy.
    shape_easy = catmull_rom_polygon(points)
    # There are some other options as well...
    shape = catmull_rom_polygon(points, subdivisions=20, extrude_height=5, show_controls=True)
    return shape_easy + right(3)(shape)

def catmull_rom_spline_variants():
    points = [
        Point2(0,0),
        Point2(1,1),
        Point2(2,1),
        Point2(2,-1),
    ]
    controls = control_points(points)

    # By default, catmull_rom_points() will return a closed smooth shape
    curve_points_closed = catmull_rom_points(points, close_loop=True)

    # If `close_loop` is False, it will return only points between the start and
    # end control points, and make a best guess about tangents for the first and last segments
    curve_points_open   = catmull_rom_points(points, close_loop=False)
    
    # By specifying start_tangent and end_tangent, you can change a shape 
    # significantly. This is similar to what you might do with Illustrator's Pen Tool.
    # Try changing these vectors to see the effects this has on the rightmost curve in the example
    start_tangent = Vector2(-2, 0)
    end_tangent = Vector2(3, 0)
    tangent_pts = [points[0] + start_tangent, *points, points[-1] + end_tangent]
    tangent_controls = control_points(tangent_pts)
    curve_points_tangents = catmull_rom_points(points, close_loop=False, 
                                start_tangent=start_tangent, end_tangent=end_tangent)

    closed = polygon(curve_points_closed) + controls
    opened = polygon(curve_points_open) + controls
    tangents = polygon(curve_points_tangents) + tangent_controls

    a = closed + right(3)(opened) + right(10)(tangents)

    return a

def catmull_rom_patches():
    # TODO: write this
    pass

def catmull_rom_prism():
    # TODO: write this
    pass

def catmull_rom_prism_smooth():
    # TODO: write this
    pass

def bottle_shape(width: float, height: float, neck_width:float=None, neck_height:float=None):
    if neck_width == None:
        neck_width = width * 0.4
    
    if neck_height == None:
        neck_height = height * 0.2

    w2 = width/2
    nw2 = neck_width/2
    h = height
    nh = neck_height

    corner_rad = 0.5

    # Add extra tangent points near curves to keep cubics from going crazy. 
    # Try taking some of these out and see how this affects the final shape
    points = [
        Point2(nw2, h),
        Point2(nw2, h-nh + 1),      # <- extra tangent
        Point2(nw2, h - nh),    
        Point2(w2, h-nh-h/6),       # <- extra tangent
        Point2(w2, corner_rad + 1), # <- extra tangent
        Point2(w2, corner_rad),
        Point2(w2-corner_rad, 0),
        Point2(0,0),
    ]
    # Use catmull_rom_points() when you don't want all corners in a polygon 
    # smoothed out or want to combine the curve with other shapes. 
    # Extra points can then be added to the list you get back
    cr_points = catmull_rom_points(points)

    # Insert a point at the top center of the bottle at the beginning of the 
    # points list. This is how the bottle has a sharp right angle corner at the 
    # sides of the neck; otherwise we'd have to insert several extra control 
    # points to make a sharp corner
    cr_points.insert(0, (0,h))
    
    # Make OpenSCAD polygons out of the shapes once all points are calculated
    a = polygon(cr_points) 
    a += mirror(v=(1,0))(a)

    # Show control points. These aren't required for anything, but seeing them
    # makes refining a curve much easier
    controls = control_points(points)
    a += controls
    return a

def basic_bezier():
    # A basic cubic Bezier curve will pass through its first and last 
    # points, but not through the central control points
    controls = [
        Point2(0, 3),
        Point2(1, 1),
        Point2(2, 1),
        Point2(3, 3)
    ]
    shape = bezier_polygon(controls, show_controls=True)
    return shape

def bezier_points_variants():
    controls = [
        Point2(0,0),
        Point2(1,2),
        Point2(2, -1),
        Point2(3,0),
    ]
    points = bezier_points(controls, subdivisions=20)
    # For non-smooth curves, add extra points
    points += [
        Point2(2, -2),
        Point2(1, -2)
    ]
    shape = polygon(points) + control_points(controls, extrude_height=0)
    return shape


if __name__ == '__main__':
    out_dir = sys.argv[1] if len(sys.argv) > 1 else os.curdir

    a = assembly()

    out_path = scad_render_to_file(a, out_dir=out_dir, include_orig_code=True)
    print(f"{__file__}: SCAD file written to: \n{out_path}")




#! /usr/bin/env python3
import sys
from pathlib import Path

from euclid3 import LineSegment2, LineSegment3, Point2, Point3

from solid import scad_render_to_file
from solid.objects import polygon, polyhedron, union
from solid.utils import forward, up

ONE_THIRD = 1 / 3


def affine_combination(a, b, weight=0.5):
    """
    Note that weight is a fraction of the distance between self and other.
    So... 0.33 is a point .33 of the way between self and other.
    """
    if hasattr(a, 'z'):
        return Point3((1 - weight) * a.x + weight * b.x,
                      (1 - weight) * a.y + weight * b.y,
                      (1 - weight) * a.z + weight * b.z,
                      )
    else:
        return Point2((1 - weight) * a.x + weight * b.x,
                      (1 - weight) * a.y + weight * b.y,
                      )


def kochify_3d(a, b, c,
               ab_weight=0.5, bc_weight=0.5, ca_weight=0.5,
               pyr_a_weight=ONE_THIRD, pyr_b_weight=ONE_THIRD, pyr_c_weight=ONE_THIRD,
               pyr_height_weight=ONE_THIRD
               ):
    """
    Point3s a, b, and c must be coplanar and define a face
    ab_weight, etc define the subdivision of the original face
    pyr_a_weight, etc define where the point of the new pyramid face will go
    pyr_height determines how far from the face the new pyramid's point will be
    """
    triangles = []
    new_a = affine_combination(a, b, ab_weight)
    new_b = affine_combination(b, c, bc_weight)
    new_c = affine_combination(c, a, ca_weight)

    triangles.extend([[a, new_a, new_c], [b, new_b, new_a], [c, new_c, new_b]])

    avg_pt_x = a.x * pyr_a_weight + b.x * pyr_b_weight + c.x * pyr_c_weight
    avg_pt_y = a.y * pyr_a_weight + b.y * pyr_b_weight + c.y * pyr_c_weight
    avg_pt_z = a.z * pyr_a_weight + b.z * pyr_b_weight + c.z * pyr_c_weight

    center_pt = Point3(avg_pt_x, avg_pt_y, avg_pt_z)

    # The top of the pyramid will be on a normal
    ab_vec = b - a
    bc_vec = c - b
    ca_vec = a - c
    normal = ab_vec.cross(bc_vec).normalized()
    avg_side_length = (abs(ab_vec) + abs(bc_vec) + abs(ca_vec)) / 3
    pyr_h = avg_side_length * pyr_height_weight
    pyr_pt = LineSegment3(center_pt, normal, pyr_h).p2

    triangles.extend([[new_a, pyr_pt, new_c], [new_b, pyr_pt, new_a], [new_c, pyr_pt, new_b]])

    return triangles


def kochify(seg, height_ratio=0.33, left_loc=0.33, midpoint_loc=0.5, right_loc=0.66):
    a, b = seg.p1, seg.p2
    l = affine_combination(a, b, left_loc)
    c = affine_combination(a, b, midpoint_loc)
    r = affine_combination(a, b, right_loc)
    # The point of the new triangle will be  height_ratio * abs(seg) long,
    # and run perpendicular to seg, through c.
    perp = seg.v.cross().normalized()

    c_height = height_ratio * abs(seg)
    perp_pt = LineSegment2(c, perp, -c_height).p2

    # For the moment, assume perp_pt is on the right side of seg.
    # Will confirm this later if needed
    return [LineSegment2(a, l),
            LineSegment2(l, perp_pt),
            LineSegment2(perp_pt, r),
            LineSegment2(r, b)]


def main_3d(out_dir):
    gens = 4

    # Parameters
    ab_weight = 0.5
    bc_weight = 0.5
    ca_weight = 0.5
    pyr_a_weight = ONE_THIRD
    pyr_b_weight = ONE_THIRD
    pyr_c_weight = ONE_THIRD
    pyr_height_weight = ONE_THIRD

    all_polys = union()

    # setup
    ax, ay, az = 100, -100, 100
    bx, by, bz = 100, 100, -100
    cx, cy, cz = -100, 100, 100
    dx, dy, dz = -100, -100, -100
    generations = [[[Point3(ax, ay, az), Point3(bx, by, bz), Point3(cx, cy, cz)],
                    [Point3(bx, by, bz), Point3(ax, ay, az), Point3(dx, dy, dz)],
                    [Point3(ax, ay, az), Point3(cx, cy, cz), Point3(dx, dy, dz)],
                    [Point3(cx, cy, cz), Point3(bx, by, bz), Point3(dx, dy, dz)],
                    ]
                   ]

    # Recursively generate snowflake segments
    for g in range(1, gens):
        generations.append([])
        for a, b, c in generations[g - 1]:
            new_tris = kochify_3d(a, b, c,
                                  ab_weight, bc_weight, ca_weight,
                                  pyr_a_weight, pyr_b_weight, pyr_c_weight,
                                  pyr_height_weight)
            generations[g].extend(new_tris)

    # Put all generations into SCAD
    orig_length = abs(generations[0][0][1] - generations[0][0][0])
    for g, a_gen in enumerate(generations):
        # Move each generation up in y so it doesn't overlap the others
        h = orig_length * 1.5 * g

        # Build the points and triangles arrays that SCAD needs
        faces = []
        points = []
        for a, b, c in a_gen:
            points.extend([[a.x, a.y, a.z], [b.x, b.y, b.z], [c.x, c.y, c.z]])
            t = len(points)
            faces.append([t - 3, t - 2, t - 1])

        # Do the SCAD
        edges = [list(range(len(points)))]
        all_polys.add(
                up(h)(
                        polyhedron(points=points, faces=faces)
                )
        )

    file_out = Path(out_dir) / 'koch_3d.scad'
    file_out = scad_render_to_file(all_polys, file_out, include_orig_code=True)
    print(f"{__file__}: SCAD file written to: {file_out}")


def main(out_dir):
    # Parameters
    height_ratio = 0.25
    left_loc = ONE_THIRD
    midpoint_loc = 0.5
    right_loc = 2 * ONE_THIRD
    gens = 5

    # Results
    all_polys = union()

    # setup
    ax, ay = 0, 0
    bx, by = 100, 0
    cx, cy = 50, 86.6
    base_seg1 = LineSegment2(Point2(ax, ay), Point2(cx, cy))
    base_seg2 = LineSegment2(Point2(cx, cy), Point2(bx, by))
    base_seg3 = LineSegment2(Point2(bx, by), Point2(ax, ay))
    generations = [[base_seg1, base_seg2, base_seg3]]

    # Recursively generate snowflake segments
    for g in range(1, gens):
        generations.append([])
        for seg in generations[g - 1]:
            generations[g].extend(kochify(seg, height_ratio, left_loc, midpoint_loc, right_loc))

    # # Put all generations into SCAD
    orig_length = abs(generations[0][0])
    for g, a_gen in enumerate(generations):
        points = [s.p1 for s in a_gen]

        # Just use arrays for points so SCAD understands
        points = [[p.x, p.y] for p in points]

        # Move each generation up in y so it doesn't overlap the others
        h = orig_length * 1.5 * g

        # Do the SCAD
        edges = [list(range(len(points)))]
        all_polys.add(forward(h)(polygon(points=points, paths=edges)))

    file_out = scad_render_to_file(all_polys, out_dir=out_dir, include_orig_code=True)
    print(f"{__file__}: SCAD file written to: {file_out}")


if __name__ == '__main__':
    out_dir = sys.argv[1] if len(sys.argv) > 1 else None
    main_3d(out_dir)
    main(out_dir)



import euclid3
from euclid3 import Vector3, Vector2, Line3

# NOTE: The PyEuclid on PyPi doesn't include several elements added to
# the module as of 13 Feb 2013.  Add them here until euclid supports them

def as_arr_local2(self):
    return [self.x, self.y]

def as_arr_local3(self):
    return [self.x, self.y, self.z]

def set_length_local2(self, length):
    d = self.magnitude()
    if d:
        factor = length / d
        self.x *= factor
        self.y *= factor

    return self

def set_length_local3(self, length):
    d = self.magnitude()
    if d:
        factor = length / d
        self.x *= factor
        self.y *= factor
        self.z *= factor

    return self

def _intersect_line3_line3(A, B):
    # Connect A & B
    # If the length of the connecting segment  is 0, they intersect
    # at the endpoint(s) of the connecting segment
    sol = euclid3._connect_line3_line3(A, B)
    # TODO: Ray3 and LineSegment3 would like to be able to know
    # if their intersection points fall within the segment.
    if sol.magnitude_squared() < 0.001:
        return sol.p
    else:
        return None

def run_euclid_patch():
    if 'as_arr' not in dir(Vector3):
        Vector3.as_arr = as_arr_local3
    if 'as_arr' not in dir(Vector2):
        Vector2.as_arr = as_arr_local2

    if 'set_length' not in dir(Vector3):
        Vector3.set_length = set_length_local3
    if 'set_length' not in dir(Vector2):
        Vector2.set_length = set_length_local2
        
    if '_intersect_line3' not in dir(Line3):
        Line3._intersect_line3 = _intersect_line3_line3



#! /usr/bin/env python3
import sys
from pathlib import Path

from solid import scad_render_to_file
from solid.objects import import_scad, use


# Import OpenSCAD code and call it from Python code.
# The path given to use() or import_scad() must be absolute or findable in sys.path


def demo_import_scad():
    scad_path = Path(__file__).parent / 'scad_to_include.scad'
    scad_mod = import_scad(scad_path)
    scad_mod.optional_nondefault_arg(1)
    return scad_mod.steps(5)


# The `use()` function mimics the behavior of OpenSCAD's use()`
def demo_scad_use():
    # scad_to_include.scad includes a module called steps()
    scad_path = Path(__file__).parent / 'scad_to_include.scad'
    # `This adds the SCAD module `steps()` to the global namespace
    use(scad_path)

    return steps(5)


if __name__ == '__main__':
    out_dir = Path(sys.argv[1]) if len(sys.argv) > 1 else None
    a = demo_import_scad()
    file_out = scad_render_to_file(a, out_dir=out_dir)
    print(f"{__file__}: SCAD file written to: \n{file_out}")



#! /usr/bin/env python3

# Basic shape with several repeated parts, demonstrating the use of
# solid.utils.bill_of_materials()
#
# Basically:
#   -- Define every part you want in the Bill of Materials in a function
#   -- Use the 'bom_part' decorator ahead of the definition of each part's function
#           e.g.:
# @bom_part()
# def my_part():
#     pass
#   -- Optionally, you can add a description and a per-unit cost to the
#       decorator invocations.
#
#   -- To generate the bill of materials, call solid.utils.bill_of_materials()
#
#       -ETJ 08 Mar 2011

import sys

from solid import scad_render_to_file
from solid.objects import cube, color, cylinder, difference, translate, union
from solid.utils import EPSILON
from solid.utils import bill_of_materials, bom_part, set_bom_headers

head_rad = 2.65
head_height = 2.8

nut_height = 2.3
nut_rad = 3

m3_rad = 1.4

doohickey_h = 5

set_bom_headers("link", "leftover")


def head():
    return cylinder(h=head_height, r=head_rad)


@bom_part("M3x16 Bolt", 0.12, currency="€", link="http://example.io/M3x16", leftover=0)
def m3_16(a=3):
    bolt_height = 16
    m = union()(
            head(),
            translate((0, 0, -bolt_height))(
                    cylinder(r=m3_rad, h=bolt_height)
            )
    )
    return m


@bom_part("M3x12 Bolt", 0.09, leftover=0)
def m3_12():
    bolt_height = 12
    m = union()(
            head(),
            translate((0, 0, -bolt_height))(
                    cylinder(r=m3_rad, h=bolt_height)
            )
    )
    return m


@bom_part("M3 Nut", 0.04, currency="R$")
def m3_nut():
    hx = cylinder(r=nut_rad, h=nut_height)
    hx.add_param('$fn', 6)  # make the nut hexagonal
    n = difference()(
            hx,
            translate((0, 0, -EPSILON))(
                    cylinder(r=m3_rad, h=nut_height + 2 * EPSILON)
            )
    )
    return n


@bom_part()
def doohickey(c):
    hole_cyl = translate((0, 0, -EPSILON))(
            cylinder(r=m3_rad, h=doohickey_h + 2 * EPSILON)
    )
    d = difference()(
            cube([30, 10, doohickey_h], center=True),
            translate((-10, 0, 0))(hole_cyl),
            hole_cyl,
            translate((10, 0, 0))(hole_cyl)
    )
    return color(c)(d)


def assembly():
    nut = m3_nut()
    return union()(
            doohickey(c='blue'),
            translate((-10, 0, doohickey_h / 2))(m3_12()),
            translate((0, 0, doohickey_h / 2))(m3_16()),
            translate((10, 0, doohickey_h / 2))(m3_12()),
            # Nuts
            translate((-10, 0, -nut_height - doohickey_h / 2))(nut),
            translate((0, 0, -nut_height - doohickey_h / 2))(nut),
            translate((10, 0, -nut_height - doohickey_h / 2))(nut),
    )


if __name__ == '__main__':
    out_dir = sys.argv[1] if len(sys.argv) > 1 else None

    a = assembly()
    bom = bill_of_materials(a)
    file_out = scad_render_to_file(a, out_dir=out_dir)

    print(f"{__file__}: SCAD file written to: \n{file_out}")
    print(bom)

    print("Or, Spreadsheet-ready TSV:\n\n")
    bom = bill_of_materials(a, csv=True)
    print(bom)



#! /usr/bin/env python3
import sys

from solid import scad_render_to_file
from solid import screw_thread
from solid.objects import cylinder

SEGMENTS = 48

inner_rad = 40
screw_height = 80


def assembly():
    section = screw_thread.default_thread_section(tooth_height=10, tooth_depth=5)
    s = screw_thread.thread(outline_pts=section,
                            inner_rad=inner_rad,
                            pitch=screw_height,
                            length=screw_height,
                            segments_per_rot=SEGMENTS,
                            neck_in_degrees=90,
                            neck_out_degrees=90)

    c = cylinder(r=inner_rad, h=screw_height)
    return s + c


if __name__ == '__main__':
    out_dir = sys.argv[1] if len(sys.argv) > 1 else None
    a = assembly()
    file_out = scad_render_to_file(a, out_dir=out_dir, include_orig_code=True)
    print(f"{__file__}: SCAD file written to: \n{file_out}")



#! /usr/bin/env python
import os
import tempfile
import unittest
from pathlib import Path

from solid.objects import background, circle, cube, cylinder, debug, disable
from solid.objects import hole, import_scad, include, part, root, rotate, sphere
from solid.objects import square, translate, use, color, difference, hull
from solid.objects import import_, intersection, intersection_for, linear_extrude, import_dxf
from solid.objects import import_stl, minkowski, mirror, multmatrix, offset, polygon
from solid.objects import polyhedron, projection, render, resize, rotate_extrude
from solid.objects import scale, surface, union

from solid.solidpython import scad_render, scad_render_animated_file, scad_render_to_file
from solid.test.ExpandedTestCase import DiffOutput

scad_test_case_templates = [
    {'name': 'polygon', 'class': 'polygon' , 'kwargs': {'paths': [[0, 1, 2]]}, 'expected': '\n\npolygon(paths = [[0, 1, 2]], points = [[0, 0], [1, 0], [0, 1]]);', 'args': {'points': [[0, 0, 0], [1, 0, 0], [0, 1, 0]]}, },
    {'name': 'polygon', 'class': 'polygon' , 'kwargs': {}, 'expected': '\n\npolygon(points = [[0, 0], [1, 0], [0, 1]]);', 'args': {'points': [[0, 0, 0], [1, 0, 0], [0, 1, 0]]}, },
    {'name': 'polygon', 'class': 'polygon' , 'kwargs': {}, 'expected': '\n\npolygon(convexity = 3, points = [[0, 0], [1, 0], [0, 1]]);', 'args': {'points': [[0, 0, 0], [1, 0, 0], [0, 1, 0]], 'convexity': 3}, },
    {'name': 'circle', 'class': 'circle' , 'kwargs': {'segments': 12, 'r': 1}, 'expected': '\n\ncircle($fn = 12, r = 1);', 'args': {}, },
    {'name': 'circle_diam', 'class': 'circle' , 'kwargs': {'segments': 12, 'd': 1}, 'expected': '\n\ncircle($fn = 12, d = 1);', 'args': {}, },
    {'name': 'square', 'class': 'square' , 'kwargs': {'center': False, 'size': 1}, 'expected': '\n\nsquare(center = false, size = 1);', 'args': {}, },
    {'name': 'sphere', 'class': 'sphere' , 'kwargs': {'segments': 12, 'r': 1}, 'expected': '\n\nsphere($fn = 12, r = 1);', 'args': {}, },
    {'name': 'sphere_diam', 'class': 'sphere' , 'kwargs': {'segments': 12, 'd': 1}, 'expected': '\n\nsphere($fn = 12, d = 1);', 'args': {}, },
    {'name': 'cube', 'class': 'cube' , 'kwargs': {'center': False, 'size': 1}, 'expected': '\n\ncube(center = false, size = 1);', 'args': {}, },
    {'name': 'cylinder', 'class': 'cylinder' , 'kwargs': {'r1': None, 'r2': None, 'h': 1, 'segments': 12, 'r': 1, 'center': False}, 'expected': '\n\ncylinder($fn = 12, center = false, h = 1, r = 1);', 'args': {}, },
    {'name': 'cylinder_d1d2', 'class': 'cylinder' , 'kwargs': {'d1': 4, 'd2': 2, 'h': 1, 'segments': 12, 'center': False}, 'expected': '\n\ncylinder($fn = 12, center = false, d1 = 4, d2 = 2, h = 1);', 'args': {}, },
    {'name': 'polyhedron', 'class': 'polyhedron' , 'kwargs': {'convexity': None}, 'expected': '\n\npolyhedron(faces = [[0, 1, 2]], points = [[0, 0, 0], [1, 0, 0], [0, 1, 0]]);', 'args': {'points': [[0, 0, 0], [1, 0, 0], [0, 1, 0]], 'faces': [[0, 1, 2]]}, },
    {'name': 'polyhedron_default_convexity', 'class': 'polyhedron' , 'kwargs': {}, 'expected': '\n\npolyhedron(convexity = 10, faces = [[0, 1, 2]], points = [[0, 0, 0], [1, 0, 0], [0, 1, 0]]);', 'args': {'points': [[0, 0, 0], [1, 0, 0], [0, 1, 0]], 'faces': [[0, 1, 2]]}, },
    {'name': 'union', 'class': 'union' , 'kwargs': {}, 'expected': '\n\nunion();', 'args': {}, },
    {'name': 'intersection', 'class': 'intersection' , 'kwargs': {}, 'expected': '\n\nintersection();', 'args': {}, },
    {'name': 'difference', 'class': 'difference' , 'kwargs': {}, 'expected': '\n\ndifference();', 'args': {}, },
    {'name': 'translate', 'class': 'translate' , 'kwargs': {'v': [1, 0, 0]}, 'expected': '\n\ntranslate(v = [1, 0, 0]);', 'args': {}, },
    {'name': 'scale', 'class': 'scale' , 'kwargs': {'v': 0.5}, 'expected': '\n\nscale(v = 0.5000000000);', 'args': {}, },
    {'name': 'rotate', 'class': 'rotate' , 'kwargs': {'a': 45, 'v': [0, 0, 1]}, 'expected': '\n\nrotate(a = 45, v = [0, 0, 1]);', 'args': {}, },
    {'name': 'mirror', 'class': 'mirror' , 'kwargs': {}, 'expected': '\n\nmirror(v = [0, 0, 1]);', 'args': {'v': [0, 0, 1]}, },
    {'name': 'resize', 'class': 'resize' , 'kwargs': {'newsize': [5, 5, 5], 'auto': [True, True, False]}, 'expected': '\n\nresize(auto = [true, true, false], newsize = [5, 5, 5]);', 'args': {}, },
    {'name': 'multmatrix', 'class': 'multmatrix' , 'kwargs': {}, 'expected': '\n\nmultmatrix(m = [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]);', 'args': {'m': [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]}, },
    {'name': 'minkowski', 'class': 'minkowski' , 'kwargs': {}, 'expected': '\n\nminkowski();', 'args': {}, },
    {'name': 'offset', 'class': 'offset' , 'kwargs': {'r': 1}, 'expected': '\n\noffset(r = 1);', 'args': {}, },
    {'name': 'offset_segments', 'class': 'offset' , 'kwargs': {'r': 1, 'segments': 12}, 'expected': '\n\noffset($fn = 12, r = 1);', 'args': {}, },
    {'name': 'offset_chamfer', 'class': 'offset' , 'kwargs': {'delta': 1}, 'expected': '\n\noffset(chamfer = false, delta = 1);', 'args': {}, },
    {'name': 'offset_zero_delta', 'class': 'offset' , 'kwargs': {'r': 0}, 'expected': '\n\noffset(r = 0);', 'args': {}, },
    {'name': 'hull', 'class': 'hull' , 'kwargs': {}, 'expected': '\n\nhull();', 'args': {}, },
    {'name': 'render', 'class': 'render' , 'kwargs': {'convexity': None}, 'expected': '\n\nrender();', 'args': {}, },
    {'name': 'projection', 'class': 'projection' , 'kwargs': {'cut': None}, 'expected': '\n\nprojection();', 'args': {}, },
    {'name': 'surface', 'class': 'surface' , 'kwargs': {'center': False, 'convexity': None}, 'expected': '\n\nsurface(center = false, file = "/Path/to/dummy.dxf");', 'args': {'file': "'/Path/to/dummy.dxf'"}, },
    {'name': 'import_stl', 'class': 'import_stl' , 'kwargs': {'layer': None, 'origin': (0, 0)}, 'expected': '\n\nimport(file = "/Path/to/dummy.stl", origin = [0, 0]);', 'args': {'file': "'/Path/to/dummy.stl'"}, },
    {'name': 'import_dxf', 'class': 'import_dxf' , 'kwargs': {'layer': None, 'origin': (0, 0)}, 'expected': '\n\nimport(file = "/Path/to/dummy.dxf", origin = [0, 0]);', 'args': {'file': "'/Path/to/dummy.dxf'"}, },
    {'name': 'import_', 'class': 'import_' , 'kwargs': {'layer': None, 'origin': (0, 0)}, 'expected': '\n\nimport(file = "/Path/to/dummy.dxf", origin = [0, 0]);', 'args': {'file': "'/Path/to/dummy.dxf'"}, },
    {'name': 'import__convexity', 'class': 'import_' , 'kwargs': {'layer': None, 'origin': (0, 0), 'convexity': 2}, 'expected': '\n\nimport(convexity = 2, file = "/Path/to/dummy.dxf", origin = [0, 0]);', 'args': {'file': "'/Path/to/dummy.dxf'"}, },
    {'name': 'linear_extrude', 'class': 'linear_extrude' , 'kwargs': {'twist': None, 'slices': None, 'center': False, 'convexity': None, 'height': 1, 'scale': 0.9}, 'expected': '\n\nlinear_extrude(center = false, height = 1, scale = 0.9000000000);', 'args': {}, },
    {'name': 'rotate_extrude', 'class': 'rotate_extrude' , 'kwargs': {'angle': 90, 'segments': 4, 'convexity': None}, 'expected': '\n\nrotate_extrude($fn = 4, angle = 90);', 'args': {}, },
    {'name': 'intersection_for', 'class': 'intersection_for' , 'kwargs': {}, 'expected': '\n\nintersection_for(n = [0, 1, 2]);', 'args': {'n': [0, 1, 2]}, },
]


class TemporaryFileBuffer(object):
    name = None
    contents = None

    def __enter__(self):
        f = tempfile.NamedTemporaryFile(delete=False)
        self.name = f.name
        try:
            f.close()
        except:
            self._cleanup()
            raise
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            with open(self.name, 'r') as f:
                self.contents = f.read()
        finally:
            self._cleanup()

    def _cleanup(self):
        try:
            os.unlink(self.name)
        except:
            pass


class TestSolidPython(DiffOutput):
    # test cases will be dynamically added to this instance

    def expand_scad_path(self, filename):
        path = Path(__file__).absolute().parent.parent / filename
        return path

    def test_infix_union(self):
        a = cube(2)
        b = sphere(2)
        expected = '\n\nunion() {\n\tcube(size = 2);\n\tsphere(r = 2);\n}'
        actual = scad_render(a + b)
        self.assertEqual(expected, actual)

    def test_infix_difference(self):
        a = cube(2)
        b = sphere(2)
        expected = '\n\ndifference() {\n\tcube(size = 2);\n\tsphere(r = 2);\n}'
        actual = scad_render(a - b)
        self.assertEqual(expected, actual)

    def test_infix_intersection(self):
        a = cube(2)
        b = sphere(2)
        expected = '\n\nintersection() {\n\tcube(size = 2);\n\tsphere(r = 2);\n}'
        actual = scad_render(a * b)
        self.assertEqual(expected, actual)

    def test_parse_scad_callables(self):
        test_str = """
                    module hex (width=10, height=10,
                                flats= true, center=false){}
                    function righty (angle=90) = 1;
                    function lefty(avar) = 2;
                    module more(a=[something, other]) {}
                    module pyramid(side=10, height=-1, square=false, centerHorizontal=true, centerVertical=false){}
                    module no_comments(arg=10,   //test comment
                    other_arg=2, /* some extra comments
                    on empty lines */
                    last_arg=4){}
                    module float_arg(arg=1.0){}
                    module arg_var(var5){}
                    module kwarg_var(var2=78){}
                    module var_true(var_true = true){}
                    module var_false(var_false = false){}
                    module var_int(var_int = 5){}
                    module var_negative(var_negative = -5){}
                    module var_float(var_float = 5.5){}
                    module var_number(var_number = -5e89){}
                    module var_empty_vector(var_empty_vector = []){}
                    module var_simple_string(var_simple_string = "simple string"){}
                    module var_complex_string(var_complex_string = "a \\"complex\\"\\tstring with a\\\\"){}
                    module var_vector(var_vector = [5454445, 565, [44545]]){}
                    module var_complex_vector(var_complex_vector = [545 + 4445, 565, [cos(75) + len("yes", 45)]]){}
                    module var_vector(var_vector = [5, 6, "string\\twith\\ttab"]){}
                    module var_range(var_range = [0:10e10]){}
                    module var_range_step(var_range_step = [-10:0.5:10]){}
                    module var_with_arithmetic(var_with_arithmetic = 8 * 9 - 1 + 89 / 15){}
                    module var_with_parentheses(var_with_parentheses = 8 * ((9 - 1) + 89) / 15){}
                    module var_with_functions(var_with_functions = abs(min(chamferHeight2, 0)) / 1){}
                    module var_with_conditional_assignment(var_with_conditional_assignment = mytest ? 45 : yop){}

                   """

        scad_file = ""
        with tempfile.NamedTemporaryFile(suffix=".scad", delete=False) as f:
            f.write(test_str.encode("utf-8"))
            scad_file = f.name

        expected = [
            {'name': 'hex', 'args': [], 'kwargs': ['width', 'height', 'flats', 'center']},
            {'name': 'righty', 'args': [], 'kwargs': ['angle']},
            {'name': 'lefty', 'args': [], 'kwargs': ['avar']},
            {'name': 'more', 'args': [], 'kwargs': ['a']},
            {'name': 'pyramid', 'args': [], 'kwargs': ['side', 'height', 'square', 'centerHorizontal', 'centerVertical']},
            {'name': 'no_comments', 'args': [], 'kwargs': ['arg', 'other_arg', 'last_arg']},
            {'name': 'float_arg', 'args': [], 'kwargs': ['arg']},
            {'name': 'arg_var', 'args': [], 'kwargs': ['var5']},
            {'name': 'kwarg_var', 'args': [], 'kwargs': ['var2']},
            {'name': 'var_true', 'args': [], 'kwargs': ['var_true']},
            {'name': 'var_false', 'args': [], 'kwargs': ['var_false']},
            {'name': 'var_int', 'args': [], 'kwargs': ['var_int']},
            {'name': 'var_negative', 'args': [], 'kwargs': ['var_negative']},
            {'name': 'var_float', 'args': [], 'kwargs': ['var_float']},
            {'name': 'var_number', 'args': [], 'kwargs': ['var_number']},
            {'name': 'var_empty_vector', 'args': [], 'kwargs': ['var_empty_vector']},
            {'name': 'var_simple_string', 'args': [], 'kwargs': ['var_simple_string']},
            {'name': 'var_complex_string', 'args': [], 'kwargs': ['var_complex_string']},
            {'name': 'var_vector', 'args': [], 'kwargs': ['var_vector']},
            {'name': 'var_complex_vector', 'args': [], 'kwargs': ['var_complex_vector']},
            {'name': 'var_vector', 'args': [], 'kwargs': ['var_vector']},
            {'name': 'var_range', 'args': [], 'kwargs': ['var_range']},
            {'name': 'var_range_step', 'args': [], 'kwargs': ['var_range_step']},
            {'name': 'var_with_arithmetic', 'args': [], 'kwargs': ['var_with_arithmetic']},
            {'name': 'var_with_parentheses', 'args': [], 'kwargs': ['var_with_parentheses']},
            {'name': 'var_with_functions', 'args': [], 'kwargs': ['var_with_functions']},
            {'name': 'var_with_conditional_assignment', 'args': [], 'kwargs': ['var_with_conditional_assignment']}
        ]

        from solid.solidpython import parse_scad_callables
        actual = parse_scad_callables(scad_file)

        for e in expected:
            self.assertEqual(e in actual, True)

        os.unlink(scad_file)

    def test_use(self):
        include_file = self.expand_scad_path("examples/scad_to_include.scad")
        use(include_file)

        a = steps(3) # type: ignore
        actual = scad_render(a)

        abs_path = a._get_include_path(include_file)
        expected = f"use <{abs_path}>\n\n\nsteps(howmany = 3);"
        self.assertEqual(expected, actual)

    def test_import_scad(self):
        include_file = self.expand_scad_path("examples/scad_to_include.scad")
        mod = import_scad(include_file)
        a = mod.steps(3)
        actual = scad_render(a)

        abs_path = a._get_include_path(include_file)
        expected = f"use <{abs_path}>\n\n\nsteps(howmany = 3);"
        self.assertEqual(expected, actual)

        # Make sure this plays nicely with `scad_render()`'s `file_header` arg
        header = '$fn = 24;'
        actual = scad_render(a, file_header=header)
        expected = f"{header}\nuse <{abs_path}>\n\n\nsteps(howmany = 3);"
        self.assertEqual(expected, actual)

        # Confirm that we can leave out even non-default arguments in OpenSCAD
        a = mod.optional_nondefault_arg();
        actual = scad_render(a)
        expected = f'use <{abs_path}>\n\n\noptional_nondefault_arg();'
        self.assertEqual(expected, actual);

        # Make sure we throw ValueError on nonexistent imports
        self.assertRaises(ValueError, import_scad, 'path/doesnt/exist.scad')

        # Test that we recursively import directories correctly
        examples = import_scad(include_file.parent)
        self.assertTrue(hasattr(examples, 'scad_to_include'))
        self.assertTrue(hasattr(examples.scad_to_include, 'steps'))

        # Test that:
        # A) scad files in the designated OpenSCAD library directories
        #       (path-dependent, see: solid.objects._openscad_library_paths())
        #       are imported correctly.
        # B) scad files in the designated app-install library directories
        from solid import objects
        lib_dirs = objects._openscad_library_paths()
        for i, ld in enumerate(lib_dirs):
            if ld.as_posix() == '.':
                continue
            if not ld.exists():
                continue
            temp_dirname = f'test_{i}'
            d = ld / temp_dirname
            d.mkdir(exist_ok=True)
            p = d / 'scad_to_include.scad'
            p.write_text(include_file.read_text())
            temp_file_str = f'{temp_dirname}/scad_to_include.scad'

            mod = import_scad(temp_file_str)
            a = mod.steps(3)
            actual = scad_render(a)
            expected = f"use <{p.absolute()}>\n\n\nsteps(howmany = 3);"
            self.assertEqual(actual, expected,
                             f'Unexpected file contents at {p} for dir: {ld}')

            # remove generated file and directories
            p.unlink()
            d.rmdir()

    def test_multiple_import_scad(self):
        # For Issue #172. Originally, multiple `import_scad()` calls would
        # re-import the entire module, rather than cache a module after one use
        include_file = self.expand_scad_path("examples/scad_to_include.scad")
        mod1 = import_scad(include_file)
        mod2 = import_scad(include_file)
        self.assertEqual(mod1, mod2)

    def test_imported_scad_arguments(self):
        include_file = self.expand_scad_path("examples/scad_to_include.scad")
        mod = import_scad(include_file)
        points = mod.scad_points();
        poly = polygon(points);
        actual = scad_render(poly);
        abs_path = points._get_include_path(include_file)
        expected = f'use <{abs_path}>\n\n\npolygon(points = scad_points());'
        self.assertEqual(expected, actual)

    def test_use_reserved_words(self):
        scad_str = '''module reserved_word_arg(or=3){\n\tcube(or);\n}\nmodule or(arg=3){\n\tcube(arg);\n}\n'''

        fd, path = tempfile.mkstemp(text=True)
        try:
            os.close(fd)
            with open(path, "w") as f:
                f.write(scad_str)

            use(path)
            a = reserved_word_arg(or_=5) # type: ignore
            actual = scad_render(a)
            expected = f"use <{path}>\n\n\nreserved_word_arg(or = 5);"
            self.assertEqual(expected, actual)

            b = or_(arg=5) # type: ignore
            actual = scad_render(b)
            expected = f"use <{path}>\n\n\nor(arg = 5);"
            self.assertEqual(expected, actual)
        finally:
            os.remove(path)

    def test_include(self):
        include_file = self.expand_scad_path("examples/scad_to_include.scad")
        self.assertIsNotNone(include_file, 'examples/scad_to_include.scad not found')
        include(include_file)
        a = steps(3) # type: ignore

        actual = scad_render(a)
        abs_path = a._get_include_path(include_file)
        expected = f"include <{abs_path}>\n\n\nsteps(howmany = 3);"
        self.assertEqual(expected, actual)

    def test_extra_args_to_included_scad(self):
        include_file = self.expand_scad_path("examples/scad_to_include.scad")
        mod = import_scad(include_file)
        a = mod.steps(3, external_var=True)
        actual = scad_render(a)

        abs_path = a._get_include_path(include_file)
        expected = f"use <{abs_path}>\n\n\nsteps(external_var = true, howmany = 3);"
        self.assertEqual(expected, actual)

    def test_background(self):
        a = cube(10)
        expected = '\n\n%cube(size = 10);'
        actual = scad_render(background(a))
        self.assertEqual(expected, actual)

    def test_debug(self):
        a = cube(10)
        expected = '\n\n#cube(size = 10);'
        actual = scad_render(debug(a))
        self.assertEqual(expected, actual)

    def test_disable(self):
        a = cube(10)
        expected = '\n\n*cube(size = 10);'
        actual = scad_render(disable(a))
        self.assertEqual(expected, actual)

    def test_root(self):
        a = cube(10)
        expected = '\n\n!cube(size = 10);'
        actual = scad_render(root(a))
        self.assertEqual(expected, actual)

    def test_color(self):
        all_args = [
            {'c': [1, 0, 0]},
            {'c': [1, 0, 0], 'alpha': 0.5},
            {'c': "#66F"},
            {'c': "Teal", 'alpha': 0.5},
        ]

        expecteds = [
            '\n\ncolor(alpha = 1.0000000000, c = [1, 0, 0]);',
            '\n\ncolor(alpha = 0.5000000000, c = [1, 0, 0]);',
            '\n\ncolor(alpha = 1.0000000000, c = "#66F");',
            '\n\ncolor(alpha = 0.5000000000, c = "Teal");',
        ]
        for args, expected in zip(all_args, expecteds):
            col = color(**args)
            actual = scad_render(col)
            self.assertEqual(expected, actual)

    def test_explicit_hole(self):
        a = cube(10, center=True) + hole()(cylinder(2, 20, center=True))
        expected = '\n\ndifference(){\n\tunion() {\n\t\tcube(center = true, size = 10);\n\t}\n\t/* Holes Below*/\n\tunion(){\n\t\tcylinder(center = true, h = 20, r = 2);\n\t} /* End Holes */ \n}'
        actual = scad_render(a)
        self.assertEqual(expected, actual)

    def test_hole_transform_propagation(self):
        # earlier versions of holes had problems where a hole
        # that was used a couple places wouldn't propagate correctly.
        # Confirm that's still happening as it's supposed to
        h = hole()(
                rotate(a=90, v=[0, 1, 0])(
                        cylinder(2, 20, center=True)
                )
        )

        h_vert = rotate(a=-90, v=[0, 1, 0])(
                h
        )

        a = cube(10, center=True) + h + h_vert
        expected = '\n\ndifference(){\n\tunion() {\n\t\tcube(center = true, size = 10);\n\t\trotate(a = -90, v = [0, 1, 0]) {\n\t\t}\n\t}\n\t/* Holes Below*/\n\tunion(){\n\t\trotate(a = 90, v = [0, 1, 0]) {\n\t\t\tcylinder(center = true, h = 20, r = 2);\n\t\t}\n\t\trotate(a = -90, v = [0, 1, 0]){\n\t\t\trotate(a = 90, v = [0, 1, 0]) {\n\t\t\t\tcylinder(center = true, h = 20, r = 2);\n\t\t\t}\n\t\t}\n\t} /* End Holes */ \n}'
        actual = scad_render(a)
        self.assertEqual(expected, actual)

    def test_separate_part_hole(self):
        # Make two parts, a block with hole, and a cylinder that
        # fits inside it.  Make them separate parts, meaning
        # holes will be defined at the level of the part_root node,
        # not the overall node.  This allows us to preserve holes as
        # first class space, but then to actually fill them in with
        # the parts intended to fit in them.
        b = cube(10, center=True)
        c = cylinder(r=2, h=12, center=True)
        p1 = b - hole()(c)

        # Mark this cube-with-hole as a separate part from the cylinder
        p1 = part()(p1)

        # This fits in the hole.  If p1 is set as a part_root, it will all appear.
        # If not, the portion of the cylinder inside the cube will not appear,
        # since it would have been removed by the hole in p1
        p2 = cylinder(r=1.5, h=14, center=True)

        a = p1 + p2

        expected = '\n\nunion() {\n\tdifference(){\n\t\tdifference() {\n\t\t\tcube(center = true, size = 10);\n\t\t}\n\t\t/* Holes Below*/\n\t\tunion(){\n\t\t\tcylinder(center = true, h = 12, r = 2);\n\t\t} /* End Holes */ \n\t}\n\tcylinder(center = true, h = 14, r = 1.5000000000);\n}'
        actual = scad_render(a)
        self.assertEqual(expected, actual)

    def test_scad_render_animated_file(self):
        def my_animate(_time=0):
            import math
            # _time will range from 0 to 1, not including 1
            rads = _time * 2 * math.pi
            rad = 15
            c = translate([rad * math.cos(rads), rad * math.sin(rads)])(square(10))
            return c

        with TemporaryFileBuffer() as tmp:
            scad_render_animated_file(my_animate, steps=2, back_and_forth=False,
                                      filepath=tmp.name, include_orig_code=False)

        actual = tmp.contents
        expected = '\nif ($t >= 0.0 && $t < 0.5){   \n\ttranslate(v = [15.0000000000, 0.0000000000]) {\n\t\tsquare(size = 10);\n\t}\n}\nif ($t >= 0.5 && $t < 1.0){   \n\ttranslate(v = [-15.0000000000, 0.0000000000]) {\n\t\tsquare(size = 10);\n\t}\n}\n'

        self.assertEqual(expected, actual)

    def test_scad_render_to_file(self):
        a = circle(10)

        # No header, no included original code
        with TemporaryFileBuffer() as tmp:
            scad_render_to_file(a, filepath=tmp.name, include_orig_code=False)

        actual = tmp.contents
        expected = '\n\ncircle(r = 10);'

        # scad_render_to_file also adds a date & version stamp before scad code;
        # That won't match here, so just make sure the correct code is at the end
        self.assertTrue(actual.endswith(expected))

        # Header
        with TemporaryFileBuffer() as tmp:
            scad_render_to_file(a, filepath=tmp.name, include_orig_code=False,
                                file_header='$fn = 24;')

        actual = tmp.contents
        expected = '$fn = 24;\n\n\ncircle(r = 10);'

        self.assertTrue(actual.endswith(expected))

        # Test out_dir specification, both using an existing dir & creating one
        # Using existing directory
        with TemporaryFileBuffer() as tmp:
            out_dir = Path(tmp.name).parent
            expected = (out_dir / 'test_solidpython.scad').as_posix()
            actual = scad_render_to_file(a, out_dir=out_dir)
            self.assertEqual(expected, actual)

        # Creating a directory on demand
        with TemporaryFileBuffer() as tmp:
            out_dir = Path(tmp.name).parent / 'SCAD'
            expected = (out_dir / 'test_solidpython.scad').as_posix()
            actual = scad_render_to_file(a, out_dir=out_dir)
            self.assertEqual(expected, actual)

        # TODO: test include_orig_code=True, but that would have to
        # be done from a separate file, or include everything in this one

    def test_numpy_type(self):
        try:
            import numpy # type: ignore
            numpy_cube = cube(size=numpy.array([1, 2, 3]))
            expected = '\n\ncube(size = [1,2,3]);'
            actual = scad_render(numpy_cube)
            self.assertEqual(expected, actual, 'Numpy SolidPython not rendered correctly')
        except ImportError:
            pass

    def test_custom_iterables(self):
        from euclid3 import Vector3

        class CustomIterable:
            def __iter__(self):
                return iter([1, 2, 3])

        expected = '\n\ncube(size = [1, 2, 3]);'
        iterables = [
            [1, 2, 3],
            (1, 2, 3),
            Vector3(1, 2, 3),
            CustomIterable(),
        ]

        for iterable in iterables:
            name = type(iterable).__name__
            actual = scad_render(cube(size=iterable))
            self.assertEqual(expected, actual, f'{name} SolidPython not rendered correctly')


def single_test(test_dict):
    name, cls, args, kwargs, expected = test_dict['name'], test_dict['class'], test_dict['args'], test_dict['kwargs'], test_dict['expected']

    def test(self):
        call_str = cls + "("
        for k, v in args.items():
            call_str += f"{k}={v}, "
        for k, v in kwargs.items():
            call_str += f"{k}={v}, "
        call_str += ')'

        scad_obj = eval(call_str)
        actual = scad_render(scad_obj)

        self.assertEqual(expected, actual)

    return test


def generate_cases_from_templates():
    for test_dict in scad_test_case_templates:
        test = single_test(test_dict)
        test_name = f"test_{test_dict['name']}"
        setattr(TestSolidPython, test_name, test)


if __name__ == '__main__':
    generate_cases_from_templates()
    unittest.main()



#! /usr/bin/env python3

#   A-Mazing Box, http://www.thingiverse.com/thing:1481
#   Copyright (C) 2009    Philipp Tiefenbacher <wizards23@gmail.com>
#   With very minor changes for SolidPython compatibility, 8 March 2011
#   With further changes for clarity, 25 September 2018
#

import os
import sys
from math import cos, pi, sin

import png

from solid import scad_render_to_file
from solid.objects import cylinder, difference, intersection, polyhedron, sphere, translate, union
from inset import insetPoly
from trianglemath import Tripple2Vec3D, angleBetweenPlanes

SEGMENTS = 48

rn = 3 * 64
# r = 10
innerR = 25
gap = 0.5
wall = 1.50
baseH = 2
gripH = 9
hn = 90
s = 0.775

h = hn * s
hone = h / hn

toph = (h - gripH) + 3


def getPNG(fn):
    with open(fn, 'rb') as f:
        r = png.Reader(file=f)
        data = r.read()
        pixel = data[2]
        raw = []
        for row in pixel:
            r = []
            raw.append(r)
            for px in row:
                r.append(px)
        return raw


def build_depth_map(img_path):
    depth = []
    for i in range(0, hn):
        depth.append([0.0] * rn)

    depth = getPNG(img_path)
    depth.reverse()
    return depth


def getPx(depth_map, x, y, default):
    x = int(x)
    y = int(y)
    x %= len(depth_map[0])
    if y >= len(depth_map):
        y = len(depth_map) - 1
    if 0 <= x < len(depth_map[0]) and 0 <= y < len(depth_map):
        return depth_map[y][x]
    return default


def myComp(x, y):
    d = Tripple2Vec3D(y).angle2D() - Tripple2Vec3D(x).angle2D()
    if d < 0:
        return -1
    elif d == 0:
        return 0
    else:
        return 1


def bumpMapCylinder(depth_map, the_r, hn_, inset, default):
    pts = []
    trls = []
    for i in range(hn_):
        circ = []
        for j in range(rn):
            a = j * 2 * pi / rn
            r = the_r - ((255 - getPx(depth_map, j, i, default)) / 150)
            p = [r * cos(a), r * sin(a), i * hone]
            circ.append(p)
        circ = insetPoly(circ, inset)
        for c in circ:
            pts.append(c)

    pts.append([0, 0, 0])
    pts.append([0, 0, i * hone])

    for j in range(rn):
        t = [j, (j + 1) % rn, rn * hn_]
        trls.append(t)
        t = [(rn * hn_ - 1) - j, (rn * hn_ - 1) - ((j + 1) % rn), rn * hn_ + 1]
        trls.append(t)
        for i in range(0, hn_ - 1):
            p1 = i * rn + ((j + 1) % rn)
            p2 = i * rn + j
            p3 = (i + 1) * rn + j
            p4 = (i + 1) * rn + ((j + 1) % rn)
            a1 = angleBetweenPlanes([pts[p1], pts[p2], pts[p3]], [pts[p4], pts[p1], pts[p3]])
            a1 = min(a1, pi - a1)
            a2 = angleBetweenPlanes([pts[p2], pts[p1], pts[p4]], [pts[p2], pts[p3], pts[p4]])
            a2 = min(a2, pi - a2)
            if a1 < a2:
                t = [p1, p2, p3]
                trls.append(t)
                t = [p4, p1, p3]
                trls.append(t)
            else:
                t = [p2, p4, p1]
                trls.append(t)
                t = [p2, p3, p4]
                trls.append(t)

    return polyhedron(pts, trls, 6)


def top_part():
    maze_path = os.path.join(os.path.dirname(__file__), 'maze7.png')

    depth_map = build_depth_map(maze_path)

    d = difference()
    u = union()
    u.add(bumpMapCylinder(depth_map, innerR, hn, 0, 255))
    u.add(cylinder(r=innerR + wall + gap, h=gripH))
    d.add(u)
    d.add(intersection()
          .add(bumpMapCylinder(depth_map, innerR, hn + 2, wall, 0).set_modifier(""))
          .add(translate((0, 0, baseH))
               .add(cylinder(r=innerR + 2 * wall, h=h * 1.1).set_modifier(""))))
    return d


def bottom_part():
    top = difference()
    u = union()
    u2 = union()
    top.add(u)
    d = difference()
    d.add(cylinder(r=innerR + wall + gap, h=toph))
    d.add(translate((0, 0, baseH)).add(cylinder(r=innerR + gap, h=toph)))
    u.add(d)
    top.add(u2)
    for i in range(0, 3):
        a = i * 2 * pi / 3
        r = innerR + gap + wall / 2
        u.add(translate(((r - 0.3) * cos(a), (r - 0.3) * sin(a), toph - 6)).add(sphere(r=2.4)))
        u2.add(translate(((r + wall - 0.3) * cos(a), (r + wall - 0.3) * sin(a), toph - 6)).add(sphere(r=2.4)))

    return top


if __name__ == '__main__':
    out_dir = sys.argv[1] if len(sys.argv) > 1 else None
    file_out = os.path.join(out_dir, 'mazebox.scad')

    assm = union()(
            top_part(),
            translate((3 * innerR, 0, 0))(
                    bottom_part()
            )
    )

    print(f"{__file__}: SCAD file written to: \n{file_out}")
    scad_render_to_file(assm, file_out, file_header=f'$fn = {SEGMENTS};')



literals = [
            ".", ",", ";",
            "=",
            "!",
            ">", "<",
            "+", "-", "*", "/", "^",
            "?", ":",
            "[", "]", "{", "}", "(", ")",
            "%", "#"
]

reserved = {
                'use'    : 'USE',
                'include': 'INCLUDE',
                'module' : 'MODULE',
                'function' : 'FUNCTION',
                'if'     : 'IF',
                'else'   : 'ELSE',
                'let'    : 'LET',
                'assert' : 'ASSERT',
                'for'    : 'FOR',
                'each'   : 'EACH',
                'true'   : 'TRUE',
                'false'  : 'FALSE',
                'echo'   : 'ECHO',
}

tokens = [
    "ID",
    "NUMBER",
    "STRING",
    "EQUAL",
    "GREATER_OR_EQUAL",
    "LESS_OR_EQUAL",
    "NOT_EQUAL",
    "AND", "OR",
    "FILENAME",
    ] + list(reserved.values())

#copy & paste from https://github.com/eliben/pycparser/blob/master/pycparser/c_lexer.py
#LICENSE: BSD
simple_escape = r"""([a-wyzA-Z._~!=&\^\-\\?'"]|x(?![0-9a-fA-F]))"""
decimal_escape = r"""(\d+)(?!\d)"""
hex_escape = r"""(x[0-9a-fA-F]+)(?![0-9a-fA-F])"""
bad_escape = r"""([\\][^a-zA-Z._~^!=&\^\-\\?'"x0-9])"""
escape_sequence = r"""(\\("""+simple_escape+'|'+decimal_escape+'|'+hex_escape+'))'
escape_sequence_start_in_string = r"""(\\[0-9a-zA-Z._~!=&\^\-\\?'"])"""
string_char = r"""([^"\\\n]|"""+escape_sequence_start_in_string+')'
t_STRING = '"'+string_char+'*"' + " | " + "'" +string_char+ "*'"

t_EQUAL = "=="
t_GREATER_OR_EQUAL = ">="
t_LESS_OR_EQUAL = "<="
t_NOT_EQUAL = "!="
t_AND = "\&\&"
t_OR = "\|\|"

t_FILENAME = r'<[a-zA-Z_0-9/\\\.-]*>'

def t_eat_escaped_quotes(t):
    r"\\\""
    pass

def t_comments1(t):
    r'(/\*(.|\n)*?\*/)'
    t.lexer.lineno += t.value.count("\n")
    pass

def t_comments2(t):
    r'//.*[\n\']?'
    t.lexer.lineno += 1
    pass

def t_whitespace(t):
    r'\s'
    t.lexer.lineno += t.value.count("\n")

def t_ID(t):
    r'[\$]?[0-9]*[a-zA-Z_][a-zA-Z_0-9]*'
    t.type = reserved.get(t.value,'ID')
    return t

def t_NUMBER(t):
    r'[0-9]*\.?\d+([eE][-\+]\d+)?'
    t.value = float(t.value)
    return t

def t_error(t):
    print(f'py_scadparser: Illegal character: {t.lexer.filename}({t.lexer.lineno}) "{t.value[0]}"')
    t.lexer.skip(1)

if __name__ == "__main__":
    import sys
    from ply import lex
    from pathlib import Path

    if len(sys.argv) < 2:
        print(f"usage: {sys.argv[0]} <scad-file>")

    p = Path(sys.argv[1])
    f = p.open()
    lexer = lex.lex()
    lexer.filename = p.as_posix()
    lexer.input(''.join(f.readlines()))
    for tok in iter(lexer.token, None):
        if tok.type == "MODULE":
            print("")
        print(repr(tok.type), repr(tok.value), end='')




from math import acos, atan2, pi, sqrt


def Tripple2Vec3D(t):
    return Vec3D(t[0], t[1], t[2])


class Vec3D:

    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    def angle2D(self):
        a = atan2(self.x, self.y)
        if a < 0:
            a += 2 * pi
        return a

    def set(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    def times(self, t):
        return Vec3D(self.x * t, self.y * t, self.z * t)

    # changes the object itself
    def add(self, v):
        self.x += v.x
        self.y += v.y
        self.z += v.z

    def plus(self, v):
        return Vec3D(self.x + v.x, self.y + v.y, self.z + v.z)

    def minus(self, v):
        return Vec3D(self.x - v.x, self.y - v.y, self.z - v.z)

    def len(self):
        return sqrt(self.x * self.x + self.y * self.y + self.z * self.z)

    # changes the object itself
    def normalize(self):
        l = self.len()
        self.x /= l
        self.y /= l
        self.z /= l

    def asTripple(self):
        return [self.x, self.y, self.z]

    def scalarProduct(self, v):
        return self.x * v.x + self.y * v.y + self.z * v.z

    def crossProduct(self, v):
        return Vec3D(self.y * v.z - self.z * v.y,
                     self.z * v.x - self.x * v.z,
                     self.x * v.y - self.y * v.x)


def planeNormal(p):
    t1 = Tripple2Vec3D(p[0])
    t2 = Tripple2Vec3D(p[1])
    t3 = Tripple2Vec3D(p[2])
    t1.add(t3.times(-1))
    t2.add(t3.times(-1))
    return t1.crossProduct(t2)


# plane defined by a list of three len 3 lists of points in R3
def angleBetweenPlanes(p1, p2):
    n1 = planeNormal(p1)
    n2 = planeNormal(p2)
    n1.normalize()
    n2.normalize()

    s = n1.scalarProduct(n2)
    if s > 1:
        s = 1
    if s < -1:
        s = -1
    return acos(s)



#! /usr/bin/env python3
import sys

from solid import scad_render_to_file
from solid.objects import cube, cylinder, difference, translate, union
from solid.utils import right

SEGMENTS = 48


def basic_geometry():
    # SolidPython code can look a lot like OpenSCAD code.  It also has
    # some syntactic sugar built in that can make it look more pythonic.
    # Here are two identical pieces of geometry, one left and one right.

    # left_piece uses standard OpenSCAD grammar (note the commas between
    # block elements; OpenSCAD doesn't require this)
    left_piece = union()(
            translate((-15, 0, 0))(
                    cube([10, 5, 3], center=True)
            ),
            translate((-10, 0, 0))(
                    difference()(
                            cylinder(r=5, h=15, center=True),
                            cylinder(r=4, h=16, center=True)
                    )
            )
    )

    # Right piece uses a more Pythonic grammar.  + (plus) is equivalent to union(), 
    # - (minus) is equivalent to difference() and * (star) is equivalent to intersection
    # solid.utils also defines up(), down(), left(), right(), forward(), and back()
    # for common transforms.
    right_piece = right(15)(cube([10, 5, 3], center=True))
    cyl = cylinder(r=5, h=15, center=True) - cylinder(r=4, h=16, center=True)
    right_piece += right(10)(cyl)

    return union()(left_piece, right_piece)


if __name__ == '__main__':
    out_dir = sys.argv[1] if len(sys.argv) > 1 else None

    a = basic_geometry()

    # Adding the file_header argument as shown allows you to change
    # the detail of arcs by changing the SEGMENTS variable.  This can
    # be expensive when making lots of small curves, but is otherwise
    # useful.
    file_out = scad_render_to_file(a, out_dir=out_dir, file_header=f'$fn = {SEGMENTS};')
    print(f"{__file__}: SCAD file written to: \n{file_out}")



# Some __init__ magic so we can include all solidpython code with:
#   from solid import *
#   from solid.utils import *
from .solidpython import scad_render, scad_render_to_file
from .solidpython import scad_render_animated, scad_render_animated_file
from .solidpython import OpenSCADObject, IncludedOpenSCADObject
from .objects import *
from .patch_euclid import run_euclid_patch

# Type hints
from .objects import P2, P3, P4, Vec3 , Vec4, Vec34, P3s, P23, Points, Indexes, ScadSize, OpenSCADObjectPlus


#! /usr/bin/env python3
import math
from typing import Sequence, Tuple, Union

from euclid3 import Point3, Vector3

from solid import scad_render_to_file
from solid.objects import cylinder, polyhedron, render
from solid.utils import EPSILON, UP_VEC, bounding_box, radians

# NOTE: The PyEuclid on PyPi doesn't include several elements added to
# the module as of 13 Feb 2013.  Add them here until euclid supports them
# TODO: when euclid updates, remove this cruft. -ETJ 13 Feb 2013
from solid import run_euclid_patch
run_euclid_patch()

P2 = Tuple[float, float]
P3 = Tuple[float, float, float]
P23 = Union[P2, P3]
Points = Sequence[P23]

def map_segment(x: float, domain_min:float, domain_max: float, range_min:float, range_max:float) -> float: 
    if domain_min == domain_max or range_min == range_max:
        return range_min
    proportion = (x - domain_min)/(domain_max - domain_min) 
    return (1-proportion) * range_min + proportion * range_max

def thread(outline_pts: Points,
           inner_rad: float,
           pitch: float,
           length: float,
           external: bool = True,
           segments_per_rot: int = 32,
           neck_in_degrees: float = 0,
           neck_out_degrees: float = 0,
           rad_2: float=None,
           inverse_thread_direction:bool=False):
    """
    Sweeps outline_pts (an array of points describing a closed polygon in XY)
    through a spiral.

    :param outline_pts: a list of points (NOT an OpenSCAD polygon) that define the cross section of the thread
    :type outline_pts: list

    :param inner_rad: radius of cylinder the screw will wrap around; at base of screw
    :type inner_rad: number

    :param pitch: height for one revolution; must be <= the height of outline_pts bounding box to avoid self-intersection
    :type pitch: number

    :param length: distance from bottom-most point of screw to topmost
    :type length: number

    :param external: if True, the cross-section is external to a cylinder. If False,the segment is internal to it, and outline_pts will be mirrored right-to-left
    :type external: bool

    :param segments_per_rot: segments per rotation
    :type segments_per_rot: int

    :param neck_in_degrees: degrees through which the outer edge of the screw thread will move from a thickness of zero (inner_rad) to its full thickness
    :type neck_in_degrees: number

    :param neck_out_degrees: degrees through which outer edge of the screw thread will move from full thickness back to zero
    :type neck_out_degrees: number

    :param rad_2: radius of cylinder the screw will wrap around at top of screw. Defaults to inner_rad
    :type rad_2: number    

    NOTE: This functions works by creating and returning one huge polyhedron, with
    potentially thousands of faces.  An alternate approach would make one single
    polyhedron,then repeat it over and over in the spiral shape, unioning them
    all together.  This would create a similar number of SCAD objects and
    operations, but still require a lot of transforms and unions to be done
    in the SCAD code rather than in the python, as here.  Also would take some
    doing to make the neck-in work as well.  Not sure how the two approaches
    compare in terms of render-time. -ETJ 16 Mar 2011

    NOTE: if pitch is less than or equal to the height of each tooth (outline_pts),
    OpenSCAD will likely crash, since the resulting screw would self-intersect
    all over the place.  For screws with essentially no space between
    threads, (i.e., pitch=tooth_height), I use pitch= tooth_height+EPSILON,
    since pitch=tooth_height will self-intersect for rotations >=1
    """
    # FIXME: For small segments_per_rot where length is not a multiple of
    # pitch, the the generated spiral will have irregularities, since we 
    # don't ensure that each level's segments are in line with those above or
    # below. This would require a change in logic to fix. For now, larger values
    # of segments_per_rot and length that divides pitch evenly should avoid this issue
    # -ETJ 02 January 2020

    rad_2 = rad_2 or inner_rad
    rotations = length / pitch

    total_angle = 360 * rotations
    up_step = length / (rotations * segments_per_rot)
    # Add one to total_steps so we have total_steps *segments*
    total_steps = math.ceil(rotations * segments_per_rot) + 1
    step_angle = total_angle / (total_steps - 1)

    all_points = []
    all_tris = []
    euc_up = Vector3(*UP_VEC)
    poly_sides = len(outline_pts)

    # Make Point3s from outline_pts and flip inward for internal threads
    int_ext_angle = 0 if external else math.pi
    outline_pts = [Point3(p[0], p[1], 0).rotate_around(axis=euc_up, theta=int_ext_angle) for p in outline_pts]

    # If this screw is conical, we'll need to rotate tooth profile to 
    # keep it perpendicular to the side of the cone. 
    if inner_rad != rad_2:
        cone_angle = -math.atan((rad_2 - inner_rad)/length) 
        outline_pts = [p.rotate_around(axis=Vector3(*UP_VEC), theta=cone_angle)  for p in outline_pts]

    # outline_pts, since they were created in 2D , are in the XY plane.
    # But spirals move a profile in XZ around the Z-axis.  So swap Y and Z
    # coordinates... and hope users know about this
    euc_points = list([Point3(p[0], 0, p[1]) for p in outline_pts])

    # Figure out how wide the tooth profile is
    min_bb, max_bb = bounding_box(outline_pts)
    outline_w = max_bb[0] - min_bb[0]
    outline_h = max_bb[1] - min_bb[1]

    # Calculate where neck-in and neck-out starts/ends
    neck_out_start = total_angle - neck_out_degrees
    neck_distance = (outline_w + EPSILON) * (1 if external else -1)
    section_rads = [
        # radius at start of thread 
        max(0, inner_rad - neck_distance),
        # end of neck-in
        map_segment(neck_in_degrees, 0, total_angle, inner_rad, rad_2),       
        # start of neck-out
        map_segment(neck_out_start, 0, total_angle, inner_rad, rad_2),
        # end of thread (& neck-out)
        rad_2 - neck_distance
    ]

    for i in range(total_steps):
        angle = i * step_angle

        elevation = i * up_step
        if angle > total_angle:
            angle = total_angle
            elevation = length

        # Handle the neck-in radius for internal and external threads
        if 0 <= angle < neck_in_degrees:
            rad = map_segment(angle, 0, neck_in_degrees, section_rads[0], section_rads[1])
        elif neck_in_degrees <= angle < neck_out_start:
            rad = map_segment( angle, neck_in_degrees, neck_out_start, section_rads[1], section_rads[2])
        elif neck_out_start <= angle <= total_angle:
            rad = map_segment( angle, neck_out_start, total_angle, section_rads[2], section_rads[3])

        elev_vec = Vector3(rad, 0, elevation)

        # create new points
        for p in euc_points:
            theta = radians(angle) * (-1 if inverse_thread_direction else 1)
            pt = (p + elev_vec).rotate_around(axis=euc_up, theta=theta)
            all_points.append(pt.as_arr())

        # Add the connectivity information
        if i < total_steps - 1:
            ind = i * poly_sides
            for j in range(ind, ind + poly_sides - 1):
                all_tris.append([j, j + 1, j + poly_sides])
                all_tris.append([j + 1, j + poly_sides + 1, j + poly_sides])
            all_tris.append([ind, ind + poly_sides - 1 + poly_sides, ind + poly_sides - 1])
            all_tris.append([ind, ind + poly_sides, ind + poly_sides - 1 + poly_sides])

    # End triangle fans for beginning and end
    last_loop = len(all_points) - poly_sides
    for i in range(poly_sides - 2):
        all_tris.append([0, i + 2, i + 1])
        all_tris.append([last_loop, last_loop + i + 1, last_loop + i + 2])

    # Moving in the opposite direction, we need to reverse the order of
    # corners in each face so the OpenSCAD preview renders correctly
    if inverse_thread_direction:
        all_tris = list([reversed(trio) for trio in all_tris])

    # Make the polyhedron; convexity info needed for correct OpenSCAD render
    a = polyhedron(points=all_points, faces=all_tris, convexity=2)

    if external:
        # Intersect with a cylindrical tube to make sure we fit into
        # the correct dimensions
        tube = cylinder(r1=inner_rad + outline_w + EPSILON, r2=rad_2 + outline_w + EPSILON, h=length, segments=segments_per_rot)
        tube -= cylinder(r1=inner_rad, r2=rad_2, h=length, segments=segments_per_rot)
    else:
        # If the threading is internal, intersect with a central cylinder
        # to make sure nothing else remains
        tube = cylinder(r1=inner_rad, r2=rad_2, h=length, segments=segments_per_rot)
    a *= tube 

    return a

def default_thread_section(tooth_height: float, tooth_depth: float):
    """
    An isosceles triangle, tooth_height vertically, tooth_depth wide:
    """
    res = [[0, -tooth_height / 2],
           [tooth_depth, 0],
           [0, tooth_height / 2]
           ]
    return res

def assembly():
    pts = [(0, -1, 0),
           (1, 0, 0),
           (0, 1, 0),
           (-1, 0, 0),
           (-1, -1, 0)]

    a = thread(pts, inner_rad=10, pitch=6, length=2, segments_per_rot=31,
               neck_in_degrees=30, neck_out_degrees=30)

    return a + cylinder(10 + EPSILON, 2)


if __name__ == '__main__':
    a = assembly()
    scad_render_to_file(a)



#! /usr/bin/env python
import difflib
from solid.solidpython import OpenSCADObject, scad_render_to_file
import unittest
import re
from euclid3 import Point3, Vector3, Point2

from solid import scad_render
from solid.objects import cube, polygon, sphere, translate
from solid.test.ExpandedTestCase import DiffOutput
from solid.utils import BoundingBox, arc, arc_inverted, euc_to_arr, euclidify 
from solid.utils import extrude_along_path, fillet_2d, is_scad, offset_points
from solid.utils import split_body_planar, transform_to_point, project_to_2D
from solid.utils import path_2d, path_2d_polygon
from solid.utils import FORWARD_VEC, RIGHT_VEC, UP_VEC
from solid.utils import back, down, forward, left, right, up
from solid.utils import label

from typing import Union

tri = [Point3(0, 0, 0), Point3(10, 0, 0), Point3(0, 10, 0)]

scad_test_cases = [
    # Test name, function, args, expected value
    ('up', up, [2], '\n\ntranslate(v = [0, 0, 2]);'),
    ('down', down, [2], '\n\ntranslate(v = [0, 0, -2]);'),
    ('left', left, [2], '\n\ntranslate(v = [-2, 0, 0]);'),
    ('right', right, [2], '\n\ntranslate(v = [2, 0, 0]);'),
    ('forward', forward, [2], '\n\ntranslate(v = [0, 2, 0]);'),
    ('back', back, [2], '\n\ntranslate(v = [0, -2, 0]);'),
    ('arc', arc, [10, 0, 90, 24], '\n\ndifference() {\n\tcircle($fn = 24, r = 10);\n\trotate(a = 0) {\n\t\ttranslate(v = [0, -10, 0]) {\n\t\t\tsquare(center = true, size = [30, 20]);\n\t\t}\n\t}\n\trotate(a = -90) {\n\t\ttranslate(v = [0, -10, 0]) {\n\t\t\tsquare(center = true, size = [30, 20]);\n\t\t}\n\t}\n}'),
    ('arc_inverted', arc_inverted, [10, 0, 90, 24], '\n\ndifference() {\n\tintersection() {\n\t\trotate(a = 0) {\n\t\t\ttranslate(v = [-990, 0, 0]) {\n\t\t\t\tsquare(center = false, size = [1000, 1000]);\n\t\t\t}\n\t\t}\n\t\trotate(a = 90) {\n\t\t\ttranslate(v = [-990, -1000, 0]) {\n\t\t\t\tsquare(center = false, size = [1000, 1000]);\n\t\t\t}\n\t\t}\n\t}\n\tcircle($fn = 24, r = 10);\n}'),
    ('transform_to_point_scad', transform_to_point, [cube(2), [2, 2, 2], [3, 3, 1]], '\n\nmultmatrix(m = [[0.7071067812, -0.1622214211, -0.6882472016, 2], [-0.7071067812, -0.1622214211, -0.6882472016, 2], [0.0000000000, 0.9733285268, -0.2294157339, 2], [0, 0, 0, 1.0000000000]]) {\n\tcube(size = 2);\n}'),
]

other_test_cases = [
    # Test name, function, args, expected value
    ('euclidify', euclidify, [[0, 0, 0]], 'Vector3(0.00, 0.00, 0.00)'),
    ('euclidify_recursive', euclidify, [[[0, 0, 0], [1, 0, 0]]], '[Vector3(0.00, 0.00, 0.00), Vector3(1.00, 0.00, 0.00)]'),
    ('euclidify_Vector', euclidify, [Vector3(0, 0, 0)], 'Vector3(0.00, 0.00, 0.00)'),
    ('euclidify_recursive_Vector', euclidify, [[Vector3(0, 0, 0), Vector3(0, 0, 1)]], '[Vector3(0.00, 0.00, 0.00), Vector3(0.00, 0.00, 1.00)]'),
    ('euclidify_3_to_2', euclidify, [Point3(0,1,2), Point2], 'Point2(0.00, 1.00)'),
    ('euc_to_arr', euc_to_arr, [Vector3(0, 0, 0)], '[0, 0, 0]'),
    ('euc_to_arr_recursive', euc_to_arr, [[Vector3(0, 0, 0), Vector3(0, 0, 1)]], '[[0, 0, 0], [0, 0, 1]]'),
    ('euc_to_arr_arr', euc_to_arr, [[0, 0, 0]], '[0, 0, 0]'),
    ('euc_to_arr_arr_recursive', euc_to_arr, [[[0, 0, 0], [1, 0, 0]]], '[[0, 0, 0], [1, 0, 0]]'),
    ('is_scad', is_scad, [cube(2)], 'True'),
    ('is_scad_false', is_scad, [2], 'False'),
    ('transform_to_point_single_arr', transform_to_point, [[1, 0, 0], [2, 2, 2], [3, 3, 1]], 'Point3(2.71, 1.29, 2.00)'),
    ('transform_to_point_single_pt3', transform_to_point, [Point3(1, 0, 0), [2, 2, 2], [3, 3, 1]], 'Point3(2.71, 1.29, 2.00)'),
    ('transform_to_point_arr_arr', transform_to_point, [[[1, 0, 0], [0, 1, 0], [0, 0, 1]], [2, 2, 2], [3, 3, 1]], '[Point3(2.71, 1.29, 2.00), Point3(1.84, 1.84, 2.97), Point3(1.31, 1.31, 1.77)]'),
    ('transform_to_point_pt3_arr', transform_to_point, [[Point3(1, 0, 0), Point3(0, 1, 0), Point3(0, 0, 1)], [2, 2, 2], [3, 3, 1]], '[Point3(2.71, 1.29, 2.00), Point3(1.84, 1.84, 2.97), Point3(1.31, 1.31, 1.77)]'),
    ('transform_to_point_redundant', transform_to_point, [[Point3(0, 0, 0), Point3(10, 0, 0), Point3(0, 10, 0)], [2, 2, 2], Vector3(0, 0, 1), Point3(0, 0, 0), Vector3(0, 1, 0), Vector3(0, 0, 1)], '[Point3(2.00, 2.00, 2.00), Point3(-8.00, 2.00, 2.00), Point3(2.00, 12.00, 2.00)]'),
    ('offset_points_inside', offset_points, [tri, 2, True], '[Point2(2.00, 2.00), Point2(5.17, 2.00), Point2(2.00, 5.17)]'),
    ('offset_points_outside', offset_points, [tri, 2, False], '[Point2(-2.00, -2.00), Point2(14.83, -2.00), Point2(-2.00, 14.83)]'),
]


class TestSPUtils(DiffOutput):
    # Test cases will be dynamically added to this instance
    # using the test case arrays above
    def assertEqualNoWhitespace(self, a, b):
        remove_whitespace = lambda s: re.subn(r'[\s\n]','', s)[0]
        self.assertEqual(remove_whitespace(a), remove_whitespace(b))

    def assertEqualOpenScadObject(self, expected:str, actual:Union[OpenSCADObject, str]):
        if isinstance(actual, OpenSCADObject):
            act = scad_render(actual)
        elif isinstance(actual, str):
            act = actual
        self.assertEqualNoWhitespace(expected, act)

    def test_split_body_planar(self):
        offset = [10, 10, 10]
        body = translate(offset)(sphere(20))
        body_bb = BoundingBox([40, 40, 40], offset)
        actual = []
        for split_dir in [RIGHT_VEC, FORWARD_VEC, UP_VEC]:
            actual_tuple = split_body_planar(body, body_bb, cutting_plane_normal=split_dir, cut_proportion=0.25)
            actual.append(actual_tuple)

        # Ignore the bounding box object that come back, taking only the SCAD objects
        actual = [scad_render(a) for splits in actual for a in splits[::2]]

        expected = ['\n\nintersection() {\n\ttranslate(v = [10, 10, 10]) {\n\t\tsphere(r = 20);\n\t}\n\ttranslate(v = [-5.0000000000, 10, 10]) {\n\t\tcube(center = true, size = [10.0000000000, 40, 40]);\n\t}\n}',
                    '\n\nintersection() {\n\ttranslate(v = [10, 10, 10]) {\n\t\tsphere(r = 20);\n\t}\n\ttranslate(v = [15.0000000000, 10, 10]) {\n\t\tcube(center = true, size = [30.0000000000, 40, 40]);\n\t}\n}',
                    '\n\nintersection() {\n\ttranslate(v = [10, 10, 10]) {\n\t\tsphere(r = 20);\n\t}\n\ttranslate(v = [10, -5.0000000000, 10]) {\n\t\tcube(center = true, size = [40, 10.0000000000, 40]);\n\t}\n}',
                    '\n\nintersection() {\n\ttranslate(v = [10, 10, 10]) {\n\t\tsphere(r = 20);\n\t}\n\ttranslate(v = [10, 15.0000000000, 10]) {\n\t\tcube(center = true, size = [40, 30.0000000000, 40]);\n\t}\n}',
                    '\n\nintersection() {\n\ttranslate(v = [10, 10, 10]) {\n\t\tsphere(r = 20);\n\t}\n\ttranslate(v = [10, 10, -5.0000000000]) {\n\t\tcube(center = true, size = [40, 40, 10.0000000000]);\n\t}\n}',
                    '\n\nintersection() {\n\ttranslate(v = [10, 10, 10]) {\n\t\tsphere(r = 20);\n\t}\n\ttranslate(v = [10, 10, 15.0000000000]) {\n\t\tcube(center = true, size = [40, 40, 30.0000000000]);\n\t}\n}'
                    ]
        self.assertEqual(actual, expected)

    def test_fillet_2d_add(self):
        pts = [[0, 5], [5, 5], [5, 0], [10, 0], [10, 10], [0, 10], ]
        p = polygon(pts)
        three_points = [euclidify(pts[0:3], Point2)]
        actual = fillet_2d(three_points, orig_poly=p, fillet_rad=2, remove_material=False)
        expected = 'union(){polygon(points=[[0,5],[5,5],[5,0],[10,0],[10,10],[0,10]]);translate(v=[3.0000000000,3.0000000000]){difference(){intersection(){rotate(a=359.9000000000){translate(v=[-998,0,0]){square(center=false,size=[1000,1000]);}}rotate(a=450.1000000000){translate(v=[-998,-1000,0]){square(center=false,size=[1000,1000]);}}}circle(r=2);}}}'
        self.assertEqualOpenScadObject(expected, actual)

    def test_fillet_2d_remove(self):
        pts = list((project_to_2D(p) for p in tri))
        poly = polygon(euc_to_arr(pts))
        actual = fillet_2d([pts], orig_poly=poly, fillet_rad=2, remove_material=True)
        expected = 'difference(){polygon(points=[[0,0],[10,0],[0,10]]);translate(v=[5.1715728753,2.0000000000]){difference(){intersection(){rotate(a=-90.1000000000){translate(v=[-998,0,0]){square(center=false,size=[1000,1000]);}}rotate(a=45.1000000000){translate(v=[-998,-1000,0]){square(center=false,size=[1000,1000]);}}}circle(r=2);}}}'
        self.assertEqualOpenScadObject(expected, actual)
    
    def test_euclidify_non_mutating(self):
        base_tri = [Point2(0, 0), Point2(10, 0), Point2(0, 10)]
        next_tri = euclidify(base_tri, Point2)
        expected = 3
        actual = len(base_tri)
        self.assertEqual(expected, actual, 'euclidify should not mutate its arguments')

    def test_offset_points_closed(self):
        actual = euc_to_arr(offset_points(tri, offset=1, closed=True))
        expected = [[1.0, 1.0], [7.585786437626904, 1.0], [1.0, 7.585786437626905]]
        self.assertEqual(expected, actual)

    def test_offset_points_open(self):
        actual = euc_to_arr(offset_points(tri, offset=1, closed=False))
        expected = [[0.0, 1.0], [7.585786437626904, 1.0], [-0.7071067811865479, 9.292893218813452]]
        self.assertEqual(expected, actual)

    def test_path_2d(self):
        base_tri = [Point2(0, 0), Point2(10, 0), Point2(10, 10)]
        actual = euc_to_arr(path_2d(base_tri, width=2, closed=False))
        expected = [
            [0.0, 1.0], [9.0, 1.0], [9.0, 10.0], 
            [11.0, 10.0], [11.0, -1.0], [0.0, -1.0]
        ]
        self.assertEqual(expected, actual)

    def test_path_2d_polygon(self):
        base_tri = [Point2(0, 0), Point2(10, 0), Point2(10, 10), Point2(0,10)]
        poly = path_2d_polygon(base_tri, width=2, closed=True)
        expected = [
            (1.0, 1.0), (9.0, 1.0), (9.0, 9.0), (1.0, 9.0), 
            (-1.0, 11.0), (11.0, 11.0), (11.0, -1.0), (-1.0, -1.0)
        ]
        actual = euc_to_arr(poly.params['points'])
        self.assertEqual(expected, actual)

        # Make sure the inner and outer paths in the polygon are disjoint
        expected = [[0,1,2,3],[4,5,6,7]]
        actual = poly.params['paths']
        self.assertEqual(expected, actual)

    def test_label(self):
        expected = 'translate(v=[0,5.0000000000,0]){resize(newsize=[15,0,0.5000000000]){union(){translate(v=[0,0.0000000000,0]){linear_extrude(height=1){text($fn=40,font="MgOpenModata:style=Bold",halign="left",spacing=1,text="Hello,",valign="baseline");}}translate(v=[0,-11.5000000000,0]){linear_extrude(height=1){text($fn=40,font="MgOpenModata:style=Bold",halign="left",spacing=1,text="World",valign="baseline");}}}}}'
        actual = label("Hello,\nWorld")
        self.assertEqualOpenScadObject(expected, actual)


def test_generator_scad(func, args, expected):
    def test_scad(self):
        scad_obj = func(*args)
        actual = scad_render(scad_obj)
        self.assertEqual(expected, actual)

    return test_scad


def test_generator_no_scad(func, args, expected):
    def test_no_scad(self):
        actual = str(func(*args))
        self.assertEqual(expected, actual)

    return test_no_scad


def read_test_tuple(test_tuple):
    if len(test_tuple) == 3:
        # If test name not supplied, create it programmatically
        func, args, expected = test_tuple
        test_name = f'test_{func.__name__}'
    elif len(test_tuple) == 4:
        test_name, func, args, expected = test_tuple
        test_name = f'test_{test_name}'
    else:
        print(f"test_tuple has {len(test_tuple):d} args :{test_tuple}")
    return test_name, func, args, expected


def create_tests():
    for test_tuple in scad_test_cases:
        test_name, func, args, expected = read_test_tuple(test_tuple)
        test = test_generator_scad(func, args, expected)
        setattr(TestSPUtils, test_name, test)

    for test_tuple in other_test_cases:
        test_name, func, args, expected = read_test_tuple(test_tuple)
        test = test_generator_no_scad(func, args, expected)
        setattr(TestSPUtils, test_name, test)


if __name__ == '__main__':
    create_tests()
    unittest.main()



#! /usr/bin/env python3
import sys

from solid import scad_render_to_file
from solid.objects import cube, cylinder, hole, part, rotate
from solid.utils import FORWARD_VEC, right, up

SEGMENTS = 120


def pipe_intersection_hole():
    pipe_od = 12
    pipe_id = 10
    seg_length = 30

    outer = cylinder(r=pipe_od, h=seg_length, center=True)
    inner = cylinder(r=pipe_id, h=seg_length + 2, center=True)

    # By declaring that the internal void of pipe_a should
    # explicitly remain empty, the combination of both pipes
    # is empty all the way through.

    # Any OpenSCAD / SolidPython object can be declared a hole(),
    # and after that will always be empty
    pipe_a = outer + hole()(inner)
    # Note that "pipe_a = outer - hole()(inner)" would work identically;
    # inner will always be subtracted now that it's a hole

    pipe_b = rotate(a=90, v=FORWARD_VEC)(pipe_a)
    return pipe_a + pipe_b


def pipe_intersection_no_hole():
    pipe_od = 12
    pipe_id = 10
    seg_length = 30

    outer = cylinder(r=pipe_od, h=seg_length, center=True)
    inner = cylinder(r=pipe_id, h=seg_length + 2, center=True)
    pipe_a = outer - inner

    pipe_b = rotate(a=90, v=FORWARD_VEC)(pipe_a)
    # pipe_a and pipe_b are both hollow, but because
    # their central voids aren't explicitly holes,
    # the union of both pipes has unwanted internal walls

    return pipe_a + pipe_b


def multipart_hole():
    # It's good to be able to keep holes empty, but often we want to put
    # things (bolts, etc.) in them.  The way to do this is to declare the
    # object containing the hole a "part".  Then, the hole will remain
    # empty no matter what you add to the 'part'.  But if you put an object
    # that is NOT part of the 'part' into the hole, it will still appear.

    # On the left (not_part), here's what happens if we try to put an object
    # into an explicit hole:  the object gets erased by the hole.

    # On the right (is_part), we mark the cube-with-hole as a "part",
    # and then insert the same 'bolt' cylinder into it.  The entire
    # bolt rematins.

    b = cube(10, center=True)
    c = cylinder(r=2, h=12, center=True)

    # A cube with an explicit hole
    not_part = b - hole()(c)

    # Mark this cube-with-hole as a separate part from the cylinder
    is_part = part()(not_part.copy())

    # This fits in the holes
    bolt = cylinder(r=1.5, h=14, center=True) + up(8)(cylinder(r=2.5, h=2.5, center=True))

    # The section of the bolt inside not_part disappears.  The section
    # of the bolt inside is_part is still there.
    return not_part + bolt + right(45)(is_part + bolt)


if __name__ == '__main__':
    out_dir = sys.argv[1] if len(sys.argv) > 1 else None

    # On the left, pipes with no explicit holes, which can give
    # unexpected walls where we don't want them.
    # On the right, we use the hole() function to fix the problem
    a = pipe_intersection_no_hole() + right(45)(pipe_intersection_hole())

    # Below is an example of how to put objects into holes and have them
    # still appear
    b = up(40)(multipart_hole())
    a += b

    file_out = scad_render_to_file(a, out_dir=out_dir, file_header=f'$fn = {SEGMENTS};', include_orig_code=True)
    print(f"{__file__}: SCAD file written to: \n{file_out}")



from math import sqrt


class Vec2D:

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def set(self, x, y):
        self.x = x
        self.y = y

    def times(self, t):
        return Vec2D(self.x * t, self.y * t)

    def add(self, v):
        self.x += v.x
        self.y += v.y

    def plus(self, v):
        return Vec2D(self.x + v.x, self.y + v.y)

    def minus(self, v):
        return Vec2D(self.x - v.x, self.y - v.y)

    def len(self):
        return sqrt(self.x * self.x + self.y * self.y)

    def normalize(self):
        l = self.len()
        self.x /= l
        self.y /= l

    def asTripple(self, z):
        return [self.x, self.y, z]

    def scalarProduct(self, v):
        return self.x * v.x + self.y * v.y

    def interpolate(self, v, t):
        return Vec2D(self.x * t + v.x * (1.0 - t), self.y * t + v.y * (1.0 - t))


class MetaCADLine:

    def __init__(self, s, e):
        self.start = Vec2D(s.x, s.y)
        self.end = Vec2D(e.x, e.y)
        self.dir = self.end.minus(self.start)
        self.normal = Vec2D(self.dir.x, self.dir.y)
        self.normal.normalize()
        self.normal.set(-self.normal.y, self.normal.x)

    def parallelMove(self, d):
        move = self.normal.times(d)
        self.start.add(move)
        self.end.add(move)

    def intersect(self, l):
        solve = LinearSolve2(l.dir.x, -self.dir.x, l.dir.y, -self.dir.y,
                             self.start.x - l.start.x, self.start.y - l.start.y)
        if solve.error:
            return None
        else:
            point = self.start.plus(self.dir.times(solve.x2))
            return Vec2D(point.x, point.y)


# matrix looks like this
# a b
# c d
def det(a, b, c, d):
    return a * d - b * c


# solves system of 2 linear equations in 2 unknown


class LinearSolve2:
    # the equations look like this looks like this
    # x1*a + x2*b = r1
    # x1*c + x2*d = r2

    def __init__(self, a, b, c, d, r1, r2):
        q = det(a, b, c, d)
        if abs(q) < 0.000000001:
            self.error = True
        else:
            self.error = False
            self.x1 = det(r1, b, r2, d) / q
            self.x2 = det(a, r1, c, r2) / q


def asVec2D(l):
    return Vec2D(l[0], l[1])


def insetPoly(poly, inset):
    points = []
    for i in range(len(poly)):
        iprev = (i + len(poly) - 1) % len(poly)
        inext = (i + 1) % len(poly)

        prev = MetaCADLine(asVec2D(poly[iprev]), asVec2D(poly[i]))
        next = MetaCADLine(asVec2D(poly[i]), asVec2D(poly[inext]))

        prev.parallelMove(inset)
        next.parallelMove(inset)

        intersect = prev.intersect(next)
        if intersect is None:
            # take parallel moved poly[i]
            # from the line thats longer (in case we    have a degenerate line
            # in there)
            if prev.dir.len() < next.dir.len():
                intersect = Vec2D(next.start.x, next.start.y)
            else:
                intersect = Vec2D(prev.end.x, prev.end.y)
        points.append(intersect.asTripple(poly[i][2]))

    istart = -1
    ilen = 0
    for i in range(len(poly)):
        iprev = (i + len(poly) - 1) % len(poly)
        prev = MetaCADLine(asVec2D(poly[iprev]), asVec2D(poly[i]))
        oldnorm = Vec2D(prev.normal.x, prev.normal.y)
        newLine = MetaCADLine(asVec2D(points[iprev]), asVec2D(points[i]))
        diff = newLine.normal.minus(oldnorm).len()
        if diff > 0.1:
            if istart == -1:
                istart = i
                ilen = 1
            else:
                ilen += 1
        else:
            if ilen > 0:
                if istart != 0:
                    idxs = (len(poly) + istart - 1) % len(poly)
                    idxe = i % len(poly)
                    p1 = points[idxs]
                    p2 = points[idxe]
                    for j in range(istart, i):
                        t = float(1 + j - istart) / (1 + i - istart)
                        points[j] = [
                            p2[0] * t + p1[0] * (1 - t),
                            p2[1] * t + p1[1] * (1 - t),
                            p2[2] * t + p1[2] * (1 - t)
                        ]
                    istart = -1
                    ilen = 0
    return points



#! /usr/bin/env python3
from solid import scad_render_to_file
from solid.objects import union

SEGMENTS = 48


def assembly():
    # Your code here!
    a = union()

    return a


if __name__ == '__main__':
    a = assembly()
    scad_render_to_file(a, file_header=f'$fn = {SEGMENTS};', include_orig_code=True)



from enum import Enum

class ScadTypes(Enum):
    GLOBAL_VAR = 0
    MODULE = 1
    FUNCTION = 2
    USE = 3
    INCLUDE = 4
    PARAMETER = 5

class ScadObject:
    def __init__(self, scadType):
        self.scadType = scadType

    def getType(self):
        return self.scadType

class ScadGlobalVar(ScadObject):
    def __init__(self, name):
        super().__init__(ScadTypes.GLOBAL_VAR)
        self.name = name

class ScadCallable(ScadObject):
    def __init__(self, name, parameters, scadType):
        super().__init__(scadType)
        self.name = name
        self.parameters = parameters

    def __repr__(self):
        return f'{self.name} ({self.parameters})'

class ScadModule(ScadCallable):
    def __init__(self, name, parameters):
        super().__init__(name, parameters, ScadTypes.MODULE)

class ScadFunction(ScadCallable):
    def __init__(self, name, parameters):
        super().__init__(name, parameters, ScadTypes.FUNCTION)

class ScadParameter(ScadObject):
    def __init__(self, name, optional=False):
        super().__init__(ScadTypes.PARAMETER)
        self.name = name
        self.optional = optional

    def __repr__(self):
        return self.name + "=None" if self.optional else  self.name




#! /usr/bin/env python

import unittest
from solid.test.ExpandedTestCase import DiffOutput
from solid import *
from solid.utils import euclidify
from solid.splines import catmull_rom_points, catmull_rom_prism, bezier_points, bezier_polygon
from euclid3 import Point2, Point3, Vector2, Vector3
from math import pi

SEGMENTS = 8

class TestSplines(DiffOutput):
    def setUp(self):
        self.points = [
            Point3(0,0),
            Point3(1,1),
            Point3(2,1),
        ]
        self.points_raw = [ (0,0), (1,1), (2,1), ]
        self.bezier_controls = [
            Point3(0,0),
            Point3(1,1),
            Point3(2,1),
            Point3(2,-1),
        ]
        self.bezier_controls_raw = [ (0,0), (1,1), (2,1), (2,-1) ]
        self.subdivisions = 2

    def assertPointsListsEqual(self, a, b):
        str_list = lambda x: list(str(v) for v in x)
        self.assertEqual(str_list(a), str_list(b))

    def test_catmull_rom_points(self):
        expected = [Point3(0.00, 0.00), Point3(0.38, 0.44), Point3(1.00, 1.00), Point3(1.62, 1.06), Point3(2.00, 1.00)]
        actual = catmull_rom_points(self.points, subdivisions=self.subdivisions, close_loop=False)
        self.assertPointsListsEqual(expected, actual)

        # TODO: verify we always have the right number of points for a given call
        # verify that `close_loop` always behaves correctly 
        # verify that start_tangent and end_tangent behavior is correct

    def test_catmull_rom_points_raw(self):
        # Verify that we can use raw sequences of floats as inputs (e.g [(1,2), (3.2,4)])
        # rather than sequences of Point2s        
        expected = [Point3(0.00, 0.00), Point3(0.38, 0.44), Point3(1.00, 1.00), Point3(1.62, 1.06), Point3(2.00, 1.00)]
        actual = catmull_rom_points(self.points_raw, subdivisions=self.subdivisions, close_loop=False)
        self.assertPointsListsEqual(expected, actual)        

    def test_catmull_rom_points_3d(self):
        points = [Point3(-1,-1,0), Point3(0,0,1), Point3(1,1,0)]
        expected = [Point3(-1.00, -1.00, 0.00), Point3(-0.62, -0.62, 0.50), Point3(0.00, 0.00, 1.00), Point3(0.62, 0.62, 0.50), Point3(1.00, 1.00, 0.00)]
        actual = catmull_rom_points(points, subdivisions=2)
        self.assertPointsListsEqual(expected, actual)

    def test_bezier_points(self):
        expected = [Point3(0.00, 0.00), Point3(1.38, 0.62), Point3(2.00, -1.00)]
        actual = bezier_points(self.bezier_controls, subdivisions=self.subdivisions)
        self.assertPointsListsEqual(expected, actual)

    def test_bezier_points_raw(self):
        # Verify that we can use raw sequences of floats as inputs (e.g [(1,2), (3.2,4)])
        # rather than sequences of Point2s
        expected = [Point3(0.00, 0.00), Point3(1.38, 0.62), Point3(2.00, -1.00)]
        actual = bezier_points(self.bezier_controls_raw, subdivisions=self.subdivisions)
        self.assertPointsListsEqual(expected, actual)   

    def test_bezier_points_3d(self):
        # verify that we get a valid bezier curve back even when its control points 
        # are outside the XY plane and aren't coplanar
        controls_3d = [Point3(-2,-1, 0), Point3(-0.5, -0.5, 1), Point3(0.5, 0.5, 1), Point3(2,1,0)]
        actual = bezier_points(controls_3d, subdivisions=self.subdivisions)
        expected = [Point3(-2.00, -1.00, 0.00),Point3(0.00, 0.00, 0.75), Point3(2.00, 1.00, 0.00)]
        self.assertPointsListsEqual(expected, actual)

    def test_catmull_rom_prism(self):
        sides = 3
        UP = Vector3(0,0,1)

        control_points = [[10, 10, 0], [10, 10, 5], [8, 8, 15]]

        cat_tube = []
        angle_step = 2*pi/sides
        for i in range(sides):
            rotated_controls = list((euclidify(p, Point3).rotate_around(UP, angle_step*i) for p in control_points))
            cat_tube.append(rotated_controls)

        poly = catmull_rom_prism(cat_tube, self.subdivisions, closed_ring=True, add_caps=True)
        actual = (len(poly.params['points']), len(poly.params['faces']))
        expected = (37, 62)
        self.assertEqual(expected, actual)
      


if __name__ == '__main__':
    unittest.main()


#! /usr/bin/env python3
from solid.objects import linear_extrude
from solid.solidpython import OpenSCADObject
import sys
from math import cos, radians, sin, pi, tau
from pathlib import Path

from euclid3 import Point2, Point3, Vector3

from solid import scad_render_to_file, text, translate, cube, color, rotate
from solid.utils import UP_VEC, Vector23, distribute_in_grid, extrude_along_path
from solid.utils import down, right, frange, lerp


from typing import Set, Sequence, List, Callable, Optional, Union, Iterable, Tuple

SEGMENTS = 48
PATH_RAD = 50 
SHAPE_RAD = 15

TEXT_LOC = [-0.6 *PATH_RAD, 1.6 * PATH_RAD]

def basic_extrude_example():
    path_rad = PATH_RAD
    shape = star(num_points=5)
    path = sinusoidal_ring(rad=path_rad, segments=240)

    # At its simplest, just sweep a shape along a path
    extruded = extrude_along_path( shape_pts=shape, path_pts=path)
    extruded += make_label('Basic Extrude')
    return extruded
        
def extrude_example_xy_scaling() -> OpenSCADObject:
    num_points = SEGMENTS
    path_rad = PATH_RAD
    circle = circle_points(15)
    path = circle_points(rad = path_rad)

    # If scales aren't included, they'll default to
    # no scaling at each step along path.
    no_scale_obj = make_label('No Scale')
    no_scale_obj += extrude_along_path(circle, path)

    # angles: from 0 to 6*Pi
    angles = list((frange(0, 3*tau, num_steps=len(path))))

    # With a 1-D scale factor, an extrusion grows and shrinks uniformly
    x_scales = [(1 + cos(a)/2) for a in angles]
    x_obj = make_label('1D Scale')
    x_obj += extrude_along_path(circle, path, scales=x_scales)

    # With a 2D scale factor, a shape's X & Y dimensions can scale 
    # independently, leading to more interesting shapes
    # X & Y scales vary between 0.5 & 1.5
    xy_scales = [Point2( 1 + cos(a)/2, 1 + sin(a)/2) for a in angles]
    xy_obj = make_label('2D Scale')
    xy_obj += extrude_along_path(circle, path, scales=xy_scales)

    obj = no_scale_obj + right(3*path_rad)(x_obj) + right(6 * path_rad)(xy_obj)
    return obj

def extrude_example_capped_ends() -> OpenSCADObject:
    num_points = SEGMENTS/2
    path_rad = 50
    circle = star(6)
    path = circle_points(rad = path_rad)[:-4]

    # If `connect_ends` is False or unspecified, ends will be capped.
    # Endcaps will be correct for most convex or mildly concave (e.g. stars) cross sections
    capped_obj = make_label('Capped Ends')
    capped_obj += extrude_along_path(circle, path, connect_ends=False, cap_ends=True)

    # If `connect_ends` is specified, create a continuous manifold object
    connected_obj = make_label('Connected Ends')
    connected_obj += extrude_along_path(circle, path, connect_ends=True)   

    return capped_obj + right(3*path_rad)(connected_obj)

def extrude_example_rotations() -> OpenSCADObject:
    path_rad = PATH_RAD
    shape = star(num_points=5)
    path = circle_points(path_rad, num_points=240)

    # For a simple example, make one complete revolution by the end of the extrusion
    simple_rot = make_label('Simple Rotation')
    simple_rot += extrude_along_path(shape, path, rotations=[360], connect_ends=True)

    # For a more complex set of rotations, add a rotation degree for each point in path
    complex_rotations = []
    degs = 0
    oscillation_max = 60

    for i in frange(0, 1, num_steps=len(path)):
        # For the first third of the path, do one complete rotation
        if i <= 0.333:
            degs = i/0.333*360
        # For the second third of the path, oscillate between +/- oscillation_max degrees 
        elif i <= 0.666:
            angle = lerp(i, 0.333, 0.666, 0, 2*tau)
            degs = oscillation_max * sin(angle)
        # For the last third of the path, oscillate increasingly fast but with smaller magnitude 
        else:
            # angle increases in a nonlinear curve, so
            # oscillations should get quicker and quicker
            x = lerp(i, 0.666, 1.0, 0, 2)
            angle = pow(x, 2.2) * tau
            # decrease the size of the oscillations by a factor of 10
            # over the course of this stretch
            osc = lerp(i, 0.666, 1.0, oscillation_max, oscillation_max/10)
            degs = osc * sin(angle)
        complex_rotations.append(degs)

    complex_rot = make_label('Complex Rotation')
    complex_rot += extrude_along_path(shape, path, rotations=complex_rotations)

    # Make some red markers to show the boundaries between the three sections of this path
    marker_w = SHAPE_RAD * 1.5
    marker = translate([path_rad, 0, 0])(
        cube([marker_w, 1, marker_w], center=True)
    )
    markers = [color('red')(rotate([0,0,120*i])(marker)) for i in range(3)]
    complex_rot += markers

    return simple_rot + right(3*path_rad)(complex_rot)

def extrude_example_transforms() -> OpenSCADObject:
    path_rad = PATH_RAD
    height = 2*SHAPE_RAD
    num_steps = 120

    shape = circle_points(rad=path_rad, num_points=120)
    path = [Point3(0,0,i) for i in frange(0, height, num_steps=num_steps)]

    max_rotation = radians(15)
    max_z_displacement = height/10
    up = Vector3(0,0,1)

    # The transforms argument is powerful. 
    # Each point in the entire extrusion will call this function with unique arguments: 
    #   -- `path_norm` in [0, 1] specifying how far along in the extrusion a point's loop is
    #   -- `loop_norm` in [0, 1] specifying where in its loop a point is.
    def point_trans(point: Point3, path_norm:float, loop_norm: float) -> Point3:
        # scale the point from 1x to 2x in the course of the 
        # extrusion, 
        scale = 1 + path_norm*path_norm/2
        p = scale * point

        # Rotate the points sinusoidally up to max_rotation
        p = p.rotate_around(up, max_rotation*sin(tau*path_norm))

        # Oscillate z values sinusoidally, growing from 
        # 0 magnitude to max_z_displacement, then decreasing to 0 magnitude at path_norm == 1
        max_z = sin(pi*path_norm) * max_z_displacement
        angle = lerp(loop_norm, 0, 1, 0, 10*tau)
        p.z += max_z*sin(angle)
        return p

    no_trans = make_label('No Transform')
    no_trans += down(height/2)(
        extrude_along_path(shape, path, cap_ends=True)
    )

    # We can pass transforms a single function that will be called on all points,
    # or pass a list with a transform function for each point along path
    arb_trans = make_label('Arbitrary Transform')
    arb_trans += down(height/2)(
        extrude_along_path(shape, path, transforms=[point_trans], cap_ends=True)
    )

    return no_trans + right(3*path_rad)(arb_trans)

# ============
# = GEOMETRY =
# ============
def sinusoidal_ring(rad=25, segments=SEGMENTS) -> List[Point3]:
    outline = []
    for i in range(segments):
        angle = radians(i * 360 / segments)
        scaled_rad = (1 + 0.18*cos(angle*5)) * rad
        x = scaled_rad * cos(angle)
        y = scaled_rad * sin(angle)
        z = 0
        # Or stir it up and add an oscillation in z as well
        # z = 3 * sin(angle * 6)
        outline.append(Point3(x, y, z))
    return outline

def star(num_points=5, outer_rad=SHAPE_RAD, dip_factor=0.5) -> List[Point3]:
    star_pts = []
    for i in range(2 * num_points):
        rad = outer_rad - i % 2 * dip_factor * outer_rad
        angle = radians(360 / (2 * num_points) * i)
        star_pts.append(Point3(rad * cos(angle), rad * sin(angle), 0))
    return star_pts

def circle_points(rad: float = SHAPE_RAD, num_points: int = SEGMENTS) -> List[Point2]:
    angles = frange(0, tau, num_steps=num_points, include_end=True)
    points = list([Point2(rad*cos(a), rad*sin(a)) for a in angles])
    return points

def make_label(message:str, text_loc:Tuple[float, float]=TEXT_LOC, height=5) -> OpenSCADObject:
    return translate(text_loc)(
        linear_extrude(height)(
            text(message)
        )
    )

# ===============
# = ENTRY POINT =
# ===============
if __name__ == "__main__":
    out_dir = sys.argv[1] if len(sys.argv) > 1 else Path(__file__).parent

    basic_extrude = basic_extrude_example()
    scaled_extrusions = extrude_example_xy_scaling()
    capped_extrusions = extrude_example_capped_ends()
    rotated_extrusions = extrude_example_rotations()
    arbitrary_transforms = extrude_example_transforms()
    all_objs = [basic_extrude, scaled_extrusions, capped_extrusions, rotated_extrusions, arbitrary_transforms]

    a = distribute_in_grid(all_objs, 
                            max_bounding_box=[4*PATH_RAD, 4*PATH_RAD],
                            rows_and_cols=[len(all_objs), 1])

    file_out = scad_render_to_file(a,  out_dir=out_dir, include_orig_code=True)
    print(f"{__file__}: SCAD file written to: \n{file_out}")



#! /usr/bin/env python3
import sys
from math import cos, sin
from typing import Optional

from solid import scad_render_animated_file
from solid.objects import square, translate
from solid.solidpython import OpenSCADObject


def my_animate(_time: Optional[float] = 0) -> OpenSCADObject:
    # _time will range from 0 to 1, not including 1
    rads = _time * 2 * 3.1416
    rad = 15
    c = translate((rad * cos(rads), rad * sin(rads)))(square(10))

    return c


if __name__ == '__main__':
    out_dir = sys.argv[1] if len(sys.argv) > 1 else None

    # To animate in OpenSCAD:
    # - Run this program to generate a SCAD file.
    # - Open the generated SCAD file in OpenSCAD
    # - Choose "View -> Animate"
    # - Enter FPS (frames per second) and Steps in the fields
    #       at the bottom of the OpenSCAD window
    # - FPS & Steps are flexible.  For a start, set both to 20 
    #       play around from there      
    file_out = scad_render_animated_file(my_animate,  # A function that takes a float argument
                                         # called '_time' in [0,1)
                                         # and returns an OpenSCAD object
                                         steps=20,  # Number of steps to create one complete motion
                                         back_and_forth=True,  # If true, runs the complete motion
                                         # forward and then in reverse,
                                         # to avoid discontinuity
                                         out_dir=out_dir,
                                         include_orig_code=True)  # Append SolidPython code
    # to the end of the generated
    # OpenSCAD code.
    print(f"{__file__}: SCAD file written to: \n{file_out}")



#! /usr/bin/env python
from math import radians
from solid import OpenSCADObject, Points, Indexes, ScadSize, polyhedron
from solid.utils import euclidify, euc_to_arr, transform_to_point, centroid
from euclid3 import Point2, Point3, Vector2, Vector3

from typing import Dict, Optional, Sequence, Tuple, Union, List, Callable

Tuple2 = Tuple[float, float]
FacetIndices = Tuple[int, int, int]
Point3Transform = Callable[[Point3, Optional[float], Optional[float]], Point3]

# ==========================
# = Extrusion along a path =
# ==========================
def extrude_along_path( shape_pts:Points, 
                        path_pts:Points, 
                        scales:Sequence[Union[Vector2, float, Tuple2]] = None,
                        rotations: Sequence[float] = None,
                        transforms: Sequence[Point3Transform] = None,
                        connect_ends = False,
                        cap_ends = True) -> OpenSCADObject:
    '''
    Extrude the curve defined by shape_pts along path_pts.
    -- For predictable results, shape_pts must be planar, convex, and lie
    in the XY plane centered around the origin. *Some* nonconvexity (e.g, star shapes)
    and nonplanarity will generally work fine
    
    -- len(scales) should equal len(path_pts).  No-op if not supplied
          Each entry may be a single number for uniform scaling, or a pair of 
          numbers (or Point2) for differential X/Y scaling
          If not supplied, no scaling will occur.
          
    -- len(rotations) should equal 1 or len(path_pts). No-op if not supplied.
          Each point in shape_pts will be rotated by rotations[i] degrees at
          each point in path_pts. Or, if only one rotation is supplied, the shape
          will be rotated smoothly over rotations[0] degrees in the course of the extrusion
    
    -- len(transforms) should be 1 or be equal to len(path_pts).  No-op if not supplied.
          Each entry should be have the signature: 
             def transform_func(p:Point3, path_norm:float, loop_norm:float): Point3
          where path_norm is in [0,1] and expresses progress through the extrusion
          and loop_norm is in [0,1] and express progress through a single loop of the extrusion
    
    -- if connect_ends is True, the first and last loops of the extrusion will
          be joined, which is useful for toroidal geometries. Overrides cap_ends

    -- if cap_ends is True, each point in the first and last loops of the extrusion
        will be connected to the centroid of that loop. For planar, convex shapes, this
        works nicely. If shape is less planar or convex, some self-intersection may happen.
        Not applied if connect_ends is True
    '''


    polyhedron_pts:Points= []
    facet_indices:List[Tuple[int, int, int]] = []

    # Make sure we've got Euclid Point3's for all elements
    shape_pts = euclidify(shape_pts, Point3)
    path_pts = euclidify(path_pts, Point3)

    src_up = Vector3(0, 0, 1)

    shape_pt_count = len(shape_pts)

    tangent_path_points: List[Point3] = []
    if connect_ends:
        tangent_path_points = [path_pts[-1]] + path_pts + [path_pts[0]]
    else:
        first = Point3(*(path_pts[0] - (path_pts[1] - path_pts[0])))
        last = Point3(*(path_pts[-1] - (path_pts[-2] - path_pts[-1])))
        tangent_path_points = [first] + path_pts + [last]
    tangents = [tangent_path_points[i+2] - tangent_path_points[i] for i in range(len(path_pts))]

    for which_loop in range(len(path_pts)):
        # path_normal is 0 at the first path_pts and 1 at the last
        path_normal = which_loop/ (len(path_pts) - 1)

        path_pt = path_pts[which_loop]
        tangent = tangents[which_loop]
        scale = scales[which_loop] if scales else 1

        rotate_degrees = None
        if rotations:
            rotate_degrees = rotations[which_loop] if len(rotations) > 1 else rotations[0] * path_normal

        transform_func = None
        if transforms:
            transform_func = transforms[which_loop] if len(transforms) > 1 else transforms[0]

        this_loop = shape_pts[:]
        this_loop = _scale_loop(this_loop, scale)
        this_loop = _rotate_loop(this_loop, rotate_degrees)
        this_loop = _transform_loop(this_loop, transform_func, path_normal)

        this_loop = transform_to_point(this_loop, dest_point=path_pt, dest_normal=tangent, src_up=src_up)
        loop_start_index = which_loop * shape_pt_count

        if (which_loop < len(path_pts) - 1):
            loop_facets = _loop_facet_indices(loop_start_index, shape_pt_count)
            facet_indices += loop_facets

        # Add the transformed points & facets to our final list
        polyhedron_pts += this_loop

    if connect_ends:
        connect_loop_start_index = len(polyhedron_pts) - shape_pt_count
        loop_facets = _loop_facet_indices(connect_loop_start_index, shape_pt_count, 0)
        facet_indices += loop_facets

    elif cap_ends:
        # OpenSCAD's polyhedron will automatically triangulate faces as needed.
        # So just include all points at each end of the tube 
        last_loop_start_index = len(polyhedron_pts) - shape_pt_count 
        start_loop_indices = list(reversed(range(shape_pt_count)))
        end_loop_indices = list(range(last_loop_start_index, last_loop_start_index + shape_pt_count))   
        facet_indices.append(start_loop_indices)
        facet_indices.append(end_loop_indices)

    return polyhedron(points=euc_to_arr(polyhedron_pts), faces=facet_indices) # type: ignore

def _loop_facet_indices(loop_start_index:int, loop_pt_count:int, next_loop_start_index=None) -> List[FacetIndices]:
    facet_indices: List[FacetIndices] = []
    # nlsi == next_loop_start_index
    if next_loop_start_index == None:
        next_loop_start_index = loop_start_index + loop_pt_count
    loop_indices      = list(range(loop_start_index,      loop_pt_count + loop_start_index)) + [loop_start_index]
    next_loop_indices = list(range(next_loop_start_index, loop_pt_count + next_loop_start_index )) + [next_loop_start_index]

    for i, (a, b) in enumerate(zip(loop_indices[:-1], loop_indices[1:])):
        c, d = next_loop_indices[i: i+2]
        # OpenSCAD's polyhedron will accept quads and do its own triangulation with them,
        # so we could just append (a,b,d,c). 
        # However, this lets OpenSCAD (Or CGAL?) do its own triangulation, leading
        # to some strange outcomes. Prefer to do our own triangulation.
        #   c--d
        #   |\ |
        #   | \|
        #   a--b               
        # facet_indices.append((a,b,d,c))
        facet_indices.append((a,b,c))
        facet_indices.append((b,d,c))
    return facet_indices

def _rotate_loop(points:Sequence[Point3], rotation_degrees:float=None) -> List[Point3]:
    if rotation_degrees is None:
        return points
    up = Vector3(0,0,1)
    rads = radians(rotation_degrees)
    return [p.rotate_around(up, rads) for p in points]

def _scale_loop(points:Sequence[Point3], scale:Union[float, Point2, Tuple2]=None) -> List[Point3]:
    if scale is None:
        return points

    if isinstance(scale, (float, int)):
        scale = [scale] * 2
    return [Point3(point.x * scale[0], point.y * scale[1], point.z) for point in points]

def _transform_loop(points:Sequence[Point3], transform_func:Point3Transform = None, path_normal:float = None) -> List[Point3]:
    # transform_func is a function that takes a point and optionally two floats,
    # a `path_normal`, in [0,1] that indicates where this loop is in a path extrusion,
    # and `loop_normal` in [0,1] that indicates where this point is in a list of points
    if transform_func is None:
        return points

    result = []
    for i, p in enumerate(points):
        # i goes from 0 to 1 across points
        loop_normal = i/(len(points) -1)
        new_p = transform_func(p, path_normal, loop_normal)
        result.append(new_p)
    return result




#! /usr/bin/env python3
import math
import random
import sys
from pathlib import Path

from solid import scad_render_to_file
from solid.objects import cube, polyhedron, translate, union


# =========================================================
# = A basic recursive Sierpinski's gasket implementation,
# outputting a file 'gasket_x.scad' to the argv[1] or $PWD
# =========================================================


class SierpinskiTetrahedron(object):

    def __init__(self, four_points):
        self.points = four_points

    def segments(self):
        indices = [(0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)]
        return [(self.points[a], self.points[b]) for a, b in indices]

    def next_gen(self, midpoint_weight=0.5, jitter_range_vec=None):
        midpoints = [weighted_midpoint(s[0], s[1], weight=midpoint_weight, jitter_range_vec=jitter_range_vec) for s in self.segments()]
        all_points = self.points + midpoints
        new_tet_indices = [(0, 4, 5, 6),
                           (4, 1, 7, 8),
                           (5, 2, 9, 7),
                           (6, 3, 8, 9), ]
        new_tets = []
        for four_ind in new_tet_indices:
            tet_points = [all_points[i] for i in four_ind]
            new_tets.append(SierpinskiTetrahedron(tet_points))

        return new_tets

    def scale(self, factor):

        self.points = [[factor * d for d in p] for p in self.points]

    def scad_code(self):
        faces = [[0, 1, 2], [0, 2, 3], [0, 3, 1], [1, 3, 2]]
        return polyhedron(points=self.points, faces=faces, convexity=1)


def distance(a, b):
    return math.sqrt((a[0] - b[0]) * (a[0] - b[0]) + (a[1] - b[1]) * (a[1] - b[1]) + (a[2] - b[2]) * (a[2] - b[2]))


def weighted_midpoint(a, b, weight=0.5, jitter_range_vec=None):
    # ignoring jitter_range_vec for now
    x = weight * a[0] + (1 - weight) * b[0]
    y = weight * a[1] + (1 - weight) * b[1]
    z = weight * a[2] + (1 - weight) * b[2]

    dist = distance(a, b)

    if jitter_range_vec:
        x += (random.random() - .5) * dist * jitter_range_vec[0]
        y += (random.random() - .5) * dist * jitter_range_vec[1]
        z += (random.random() - .5) * dist * jitter_range_vec[2]

    return [x, y, z]


def sierpinski_3d(generation, scale=1, midpoint_weight=0.5, jitter_range_vec=None):
    orig_tet = SierpinskiTetrahedron([[1.0, 1.0, 1.0],
                                      [-1.0, -1.0, 1.0],
                                      [-1.0, 1.0, -1.0],
                                      [1.0, -1.0, -1.0]])
    all_tets = [orig_tet]
    for i in range(generation):
        all_tets = [subtet for tet in all_tets for subtet in tet.next_gen(midpoint_weight, jitter_range_vec)]

    if scale != 1:
        for tet in all_tets:
            tet.scale(scale)
    return all_tets


if __name__ == '__main__':
    out_dir = Path(sys.argv[1]) if len(sys.argv) > 1 else Path.cwd()

    generations = 3
    midpoint_weight = 0.5
    # jitter_range_vec adds some randomness to the generated shape,
    # making it more interesting.  Try:
    # jitter_range_vec = [0.5,0, 0]
    jitter_range_vec = None
    all_tets = sierpinski_3d(generations, scale=100, midpoint_weight=midpoint_weight, jitter_range_vec=jitter_range_vec)

    t = union()
    for tet in all_tets:
        # Create the scad code for all tetrahedra
        t.add(tet.scad_code())
        # Draw cubes at all intersections to make the shape manifold.
        for p in tet.points:
            t.add(translate(p).add(cube(5, center=True)))

    file_out = out_dir / f'gasket_{generations}_gen.scad'
    file_out = scad_render_to_file(t, file_out)
    print(f"{__file__}: SCAD file written to: \n{file_out}")



#! /usr/bin/env python3
import sys

from solid import scad_render_to_file
from solid.objects import cylinder
from solid.utils import up

SEGMENTS = 48


def show_appended_python_code():
    a = cylinder(r=10, h=10, center=True) + up(5)(cylinder(r1=10, r2=0, h=10))

    return a


if __name__ == '__main__':
    out_dir = sys.argv[1] if len(sys.argv) > 1 else None
    a = show_appended_python_code()

    # ================================================================
    # = include_orig_code appends all python code as comments to the
    # = bottom of the generated OpenSCAD code, so the final document
    # = contains the easy-to-read python code as well as the SCAD.
    # = ------------------------------------------------------------ =
    file_out = scad_render_to_file(a, out_dir=out_dir, include_orig_code=True)
    print(f"{__file__}: SCAD file written to: \n{file_out}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['solid',
 'solid.examples',
 'solid.examples.mazebox',
 'solid.mypy',
 'solid.py_scadparser',
 'solid.test']

package_data = \
{'': ['*'],
 'solid': ['.mypy_cache/3.7/*',
           '.mypy_cache/3.7/collections/*',
           '.mypy_cache/3.7/importlib/*',
           '.mypy_cache/3.7/os/*',
           '.mypy_cache/3.7/solid/*',
           '.vscode/*'],
 'solid.examples': ['.mypy_cache/3.7/*',
                    '.mypy_cache/3.7/collections/*',
                    '.mypy_cache/3.7/importlib/*',
                    '.mypy_cache/3.7/os/*',
                    'Compiled_examples/*'],
 'solid.mypy': ['.mypy_cache/3.7/*',
                '.mypy_cache/3.7/collections/*',
                '.mypy_cache/3.7/importlib/*',
                '.mypy_cache/3.7/os/*'],
 'solid.test': ['solidpython.egg-info/*', 'test_0/*']}

install_requires = \
['PrettyTable==0.7.2',
 'euclid3>=0.1.0,<0.2.0',
 'ply>=3.11,<4.0',
 'pypng>=0.0.19,<0.0.20']

setup_kwargs = {
    'name': 'solidpython',
    'version': '1.1.3',
    'description': 'Python interface to the OpenSCAD declarative geometry language',
    'long_description': None,
    'author': 'Evan Jones',
    'author_email': 'evan_t_jones@mac.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SolidCode/SolidPython',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)
